#  Copyright 2019 Francesco Ceccon
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Bounds tightening of univariate quadratic expression"""
from collections import namedtuple
from pyomo.core.kernel.component_map import ComponentMap
from suspect.math import (
    almosteq, isinf, inf, sqrt, RoundMode as RM
)
from suspect.expression import ExpressionType
from suspect.interval import Interval
from suspect.interfaces import Rule


_VariableCoefficients = \
    namedtuple('_VariableCoefficients', ['quadratic', 'linear'])


class UnivariateQuadraticRule(Rule):
    """Bounds tightening of univariate quadratic expression.

    References
    ----------
    [1] Decomposition of multistage stochastic programs and a constraint integer
    programming approach to mixed-integer nonlinear programming. S. Vigerske.
    PhD thesis, Humboldt Universität zu Berlin
    """
    def __init__(self):
        pass

    def apply(self, expr, bounds):
        # Look for Quadratic + Linear expressions
        if expr.nargs() != 2:
            return

        quadratic, linear = _quadratic_and_linear(expr.children)

        if quadratic is None or linear is None:
            return

        univariate_terms, quadratic_terms, linear_terms = \
            _collect_expression_types(quadratic, linear)

        # Compute bounds of non univariate terms. We will use this to perform
        # tightening on univariate terms.
        quadratic_terms_bound = sum(
            t.coefficient * bounds[t.var1] * bounds[t.var2]
            for t in quadratic_terms
        )
        linear_terms_bound = sum(
            bounds[v] * c
            for v, c in linear_terms
        )
        expr_bound = bounds.get(expr, Interval(None, None))

        non_univariate_bound = \
            expr_bound - quadratic_terms_bound - linear_terms_bound

        univariate_terms_bounds = [
            _univariate_term_bound(a, b, bounds[v])
            for (v, a, b) in univariate_terms
        ]

        tightened_univariate_terms_bounds = [
            _tighten_univariate_term_bound(
                term_idx, univariate_terms_bounds, non_univariate_bound
            )
            for term_idx, _ in enumerate(univariate_terms_bounds)
        ]

        vars_bounds = ComponentMap(
            (v, _variable_bound_from_univariate_bound(a, b, bound))
            for bound, (v, a, b)
            in zip(tightened_univariate_terms_bounds, univariate_terms)
        )

        return vars_bounds


def _variable_bound_from_univariate_bound(a, b, bound):
    bounds_from_lower_bound = _variable_bound_from_univariate_lower_bound(
        a, b, bound.lower_bound
    )
    bounds_from_upper_bound = _variable_bound_from_univariate_lower_bound(
        -a, -b, -bound.upper_bound
    )
    return bounds_from_lower_bound.intersect(bounds_from_upper_bound)


def _variable_bound_from_univariate_lower_bound(a, b, c):
    t = c/a + (b**2) / (4*a**2)

    if a > 0:
        return Interval(None, None)
        # if c + (b**2) / (4*a) <= 0:
        #     return Interval(None, None)
        # new_upper_bound = Interval(None, -sqrt(t, RM.RN) - b/(2*a))
        # new_lower_bound = Interval(sqrt(t, RM.RN) - b/(2*a), None)
        # return new_upper_bound.intersect(new_lower_bound)

    # a < 0
    if c + (b**2) / (4*a) > 0:
        # The bound would be empty which is something that shouldn't happen.
        # Return [-inf, inf] so we don't tighten the bounds
        return Interval(None, None)

    return Interval(-sqrt(t, RM.RN) - b/(2*a), sqrt(t, RM.RN) - b/(2*a))


def _tighten_univariate_term_bound(term_idx, univariate_terms_bounds,
                                   expr_bounds):
    siblings_bound = sum(
        bound
        for i, bound in enumerate(univariate_terms_bounds)
        if i != term_idx
    )
    return expr_bounds - siblings_bound


def _univariate_term_bound(a, b, x_bound):
    """Return the bound of the univariate term ax^2 + bx"""
    lower_bound = x_bound.lower_bound
    upper_bound = x_bound.upper_bound

    if lower_bound is None or isinf(lower_bound):
        univariate_lower_bound = -inf
    else:
        univariate_lower_bound = a * lower_bound**2 + b*lower_bound

    if upper_bound is None or isinf(upper_bound):
        univariate_upper_bound = inf
    else:
        univariate_upper_bound = a * upper_bound**2 + b*upper_bound

    if -b/(2*a) in x_bound:
        t = -(b**2) / (4*a)
        new_lower_bound = min(univariate_lower_bound, univariate_upper_bound, t)
        new_upper_bound = max(univariate_lower_bound, univariate_upper_bound, t)
        return Interval(new_lower_bound, new_upper_bound)
    new_lower_bound = min(univariate_lower_bound, univariate_upper_bound)
    new_upper_bound = max(univariate_lower_bound, univariate_upper_bound)
    return Interval(new_lower_bound, new_upper_bound)


def _collect_expression_types(quadratic, linear):
    """Collect different expression types from quadratic and linear expression.

    Given an expression

    a0 x0^2 + a1 x1^2  + ...+ b0 x0 + b1 x1 + b2 x2 + ...

    Returns the quadratic univariate expressions like `a0 x0^2`, the bilinear
    expressions `c x1 x2`, and the linear expressions `bn xn`.

    :param quadratic: the quadratic expression
    :param linear: the linear expression
    :return:
    A tuple (univariate_expr, bilinear_terms, linear_terms)
    """
    variables_coef = ComponentMap()

    for coef, var in zip(linear.linear_coefs, linear.linear_vars):
        variables_coef[var] = \
            _VariableCoefficients(quadratic=0.0, linear=coef)

    non_univariate_terms = []
    for term in quadratic.terms:
        if term.var1 is term.var2:
            var_coef = variables_coef.get(term.var1, None)
            if var_coef is None:
                non_univariate_terms.append(term)
            else:
                linear_coef = var_coef.linear
                var_coef = _VariableCoefficients(
                    quadratic=term.coefficient, linear=linear_coef
                )
                variables_coef[term.var1] = var_coef
        else:
            non_univariate_terms.append(term)

    linear_terms = [
        (v, vc.linear)
        for v, vc in variables_coef.items()
        if almosteq(vc.quadratic, 0.0)
    ]
    univariate_terms = [
        (v, vc.quadratic, vc.linear)
        for v, vc in variables_coef.items()
        if not almosteq(vc.quadratic, 0.0)
    ]

    return univariate_terms, non_univariate_terms, linear_terms


def _quadratic_and_linear(children):
    assert len(children) == 2
    a, b = children
    if a.expression_type == ExpressionType.Quadratic:
        if b.expression_type == ExpressionType.Linear:
            return a, b
        return None, None

    if b.expression_type == ExpressionType.Quadratic:
        if a.expression_type == ExpressionType.Linear:
            return b, a

    return None, None



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convexity detector for fractional expressions."""
from suspect.ext import ConvexityDetector
from suspect.convexity import Convexity
from suspect.interfaces import Rule
from suspect.expression import ExpressionType
from suspect.math import almosteq # pylint: disable=no-name-in-module


class FractionalConvexityDetector(ConvexityDetector):
    """Convexity detector for fractional expressions of the type:

        g(x) = (a1*x + b1) / (a2*x + b2)

    the convexity is derived from the second derivative of the expression:

        ddg/ddx = -2*a2*(a1*b2 - a2*b1)/(a1*x + b2)^3
    """
    def __init__(self, _problem):
        super().__init__()
        self._rule = FractionalRule()

    def visit_expression(self, expr, convexity, mono, bounds):
        return False, None


class FractionalRule(Rule):
    """Convexity detector for fractional expressions."""
    linear_types = (
        ExpressionType.Constant,
        ExpressionType.Variable,
        ExpressionType.Linear,
    )

    def apply(self, expr, ctx):
        num, den = expr.children
        if not self._valid_expression(num, den):
            return None

        a1, x, b1 = self._linear_components(num)
        a2, y, b2 = self._linear_components(den)

        if x is not None and y is not None:
            if x is not y:
                return None

        if x is not None:
            x_bound = ctx.bounds(x)
        elif y is not None:
            x_bound = ctx.bounds(y)
        else:
            raise RuntimeError('no variable in numerator or denonimator')

        dd_num = -2*a2*(a1*b2 - a2*b1)
        dd_den = a2*x_bound + b2

        if almosteq(dd_num, 0):
            return Convexity.Linear

        dd_bound = dd_den * (1/dd_num)

        if dd_bound.is_nonpositive():
            return Convexity.Concave
        elif dd_bound.is_nonnegative():
            return Convexity.Convex
        # inconclusive
        return None

    def _valid_expression(self, num, den):
        num_type = num.expression_type
        den_type = den.expression_type
        if num_type not in self.linear_types:
            return False
        if den_type not in self.linear_types:
            return False
        if num_type == ExpressionType.Linear:
            if len(num.children) != 1:
                return False
        if den_type == ExpressionType.Linear:
            if len(den.children) != 1:
                return False
        return True

    def _linear_components(self, expr):
        expr_type = expr.expression_type
        if expr_type == ExpressionType.Linear:
            assert len(expr.children) == 1
            x = expr.children[0]
            a = expr.coefficient(x)
            b = expr.constant_term
            return a, x, b
        if expr_type == ExpressionType.Constant:
            return 0.0, None, expr.value
        if expr_type == ExpressionType.Variable:
            return 1.0, expr, 0.0
        raise RuntimeError('invalid expression type in fractional convexity detector')



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convexity detection rules for trigonometric expressions."""
from suspect.convexity.convexity import Convexity
from suspect.convexity.rules.rule import ConvexityRule
from suspect.interval import Interval
from suspect.math import pi # pylint: disable=no-name-in-module


class SinRule(ConvexityRule):
    """Return convexity of sin."""
    def apply(self, expr, convexity, _mono, bounds): # pylint: disable=too-many-return-statements
        child = expr.args[0]
        child_bounds = bounds[child]
        cvx = convexity[child]

        if child_bounds.size() > pi:
            return Convexity.Unknown

        sin_bounds = child_bounds.sin()
        if sin_bounds.lower_bound * sin_bounds.upper_bound < 0:
            return Convexity.Unknown

        cos_bounds = child_bounds.cos()

        if sin_bounds.is_nonnegative():
            if cvx.is_linear():
                return Convexity.Concave
            if cvx.is_convex() and cos_bounds.is_nonpositive():
                return Convexity.Concave
            if cvx.is_concave() and cos_bounds.is_nonnegative():
                return Convexity.Concave
        elif sin_bounds.is_nonpositive():
            if cvx.is_linear():
                return Convexity.Convex
            if cvx.is_concave() and cos_bounds.is_nonpositive():
                return Convexity.Convex
            if cvx.is_convex() and cos_bounds.is_nonnegative():
                return Convexity.Convex

        return Convexity.Unknown


class AsinRule(ConvexityRule):
    """Return convexity of asin."""
    def apply(self, expr, convexity, _mono, bounds):
        child = expr.args[0]
        child_bounds = bounds[child]
        cvx = convexity[child]

        concave_domain = Interval(-1, 0)
        if child_bounds in concave_domain and cvx.is_concave():
            return Convexity.Concave

        convex_domain = Interval(0, 1)
        if child_bounds in convex_domain and cvx.is_convex():
            return Convexity.Convex

        return Convexity.Unknown



class CosRule(ConvexityRule):
    """Return convexity of cos."""
    def apply(self, expr, convexity, _mono, bounds): # pylint: disable=too-many-return-statements
        child = expr.args[0]
        child_bounds = bounds[child]
        cvx = convexity[child]

        if child_bounds.size() > pi:
            return Convexity.Unknown

        cos_bounds = child_bounds.cos()
        if cos_bounds.lower_bound * cos_bounds.upper_bound < 0:
            return Convexity.Unknown

        sin_bounds = child_bounds.sin()

        if cos_bounds.is_nonnegative():
            if cvx.is_linear():
                return Convexity.Concave
            if cvx.is_convex() and sin_bounds.is_nonpositive():
                return Convexity.Concave
            if cvx.is_concave() and sin_bounds.is_nonnegative():
                return Convexity.Concave
        elif cos_bounds.is_nonpositive():
            if cvx.is_linear():
                return Convexity.Convex
            if cvx.is_concave() and sin_bounds.is_nonpositive():
                return Convexity.Convex
            if cvx.is_convex() and sin_bounds.is_nonnegative():
                return Convexity.Convex

        return Convexity.Unknown


class AcosRule(ConvexityRule):
    """Return convexity of acos."""
    def apply(self, expr, convexity, _mono, bounds):
        child = expr.args[0]
        child_bounds = bounds[child]
        cvx = convexity[child]

        convex_domain = Interval(-1, 0)
        if child_bounds in convex_domain and cvx.is_concave():
            return Convexity.Convex

        concave_domain = Interval(0, 1)
        if child_bounds in concave_domain and cvx.is_convex():
            return Convexity.Concave

        return Convexity.Unknown


class TanRule(ConvexityRule):
    """Return convexity of tan."""
    def apply(self, expr, convexity, _mono, bounds):
        child = expr.args[0]
        child_bounds = bounds[child]
        cvx = convexity[child]

        if 2.0*child_bounds.size() > pi:
            return Convexity.Unknown

        tan_bounds = child_bounds.tan()
        if tan_bounds.lower_bound * tan_bounds.upper_bound < 0:
            return Convexity.Unknown

        if tan_bounds.is_nonnegative() and cvx.is_convex():
            return Convexity.Convex

        if tan_bounds.is_nonpositive() and cvx.is_concave():
            return Convexity.Concave

        return Convexity.Unknown


class AtanRule(ConvexityRule):
    """Return convexity of atan."""
    def apply(self, expr, convexity, _mono, bounds):
        child = expr.args[0]
        child_bounds = bounds[child]
        cvx = convexity[child]

        if child_bounds.is_nonpositive() and cvx.is_convex():
            return Convexity.Convex

        if child_bounds.is_nonnegative() and cvx.is_concave():
            return Convexity.Concave

        return Convexity.Unknown



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing VerticesList."""
import bisect


# pylint: disable=invalid-name
def _reverse_bisect_right(arr, x):
    """Like bisect.bisect_right, but insert in a descending list"""
    lo = 0
    hi = len(arr)
    while lo < hi:
        mid = (lo + hi) // 2
        if x > arr[mid]:
            hi = mid
        else:
            lo = mid + 1
    return lo


class VerticesList(object):
    """A list of vertices sorted by their depth."""
    def __init__(self, reverse=False):
        self._vertices = []
        self._vertices_depth = []
        self._vertices_set = set()
        self._reverse = reverse

        if self._reverse:
            self._find_insertion_idx = _reverse_bisect_right
        else:
            self._find_insertion_idx = bisect.bisect_right

    def append(self, vertex, depth):
        """Append vertex to the list, keeping the vertices sorted by depth"""
        if id(vertex) in self._vertices_set:
            return
        insertion_idx = self._find_insertion_idx(self._vertices_depth, depth)
        self._vertices.insert(insertion_idx, vertex)
        self._vertices_depth.insert(insertion_idx, depth)
        self._vertices_set.add(id(vertex))

    def pop(self):
        """Pop an element from the front of the list"""
        self._vertices_depth.pop(0)
        vertex = self._vertices.pop(0)
        if id(vertex) in self._vertices_set:
            self._vertices_set.remove(id(vertex))
        return vertex

    def __iter__(self):
        return iter(self._vertices)

    def __len__(self):
        return len(self._vertices)

    def __contains__(self, vertex):
        return id(vertex) in self._vertices_set

    def __bool__(self):
        return len(self) > 0



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convexity detector for fractional expressions."""
import numpy as np
from suspect.ext import ConvexityDetector
from suspect.convexity import Convexity
from suspect.interfaces import Rule
from suspect.expression import ExpressionType, UnaryFunctionType


class PerspectiveFunctionConvexityDetector(ConvexityDetector):
    """Convexity detector for perspective functions."""
    def __init__(self, _problem):
        super().__init__()

    def visit_expression(self, expr, convexity, mono, bounds):
        return False, None
        return {
            ExpressionType.Product: PerspectiveRule(),
        }


class PerspectiveRule(Rule):
    """Convexity detector for perspective functions."""
    root_expr = ExpressionType.Product

    @property
    def _convex_concave_functions(self):
        return self._convex_functions + self._concave_functions

    @property
    def _convex_functions(self):
        return [UnaryFunctionType.Exp]

    @property
    def _concave_functions(self):
        return [UnaryFunctionType.Sqrt, UnaryFunctionType.Log]

    def apply(self, expr, ctx):
        f, g = expr.children
        f_type = f.expression_type
        g_type = g.expression_type
        if f_type == ExpressionType.Linear and g_type == ExpressionType.Sum:
            return self._detect_syn_convexity_outer(f, g, ctx)
        elif g_type == ExpressionType.Linear and f_type == ExpressionType.Sum:
            return self._detect_syn_convexity_outer(g, f, ctx)
        elif f_type == ExpressionType.Linear and g_type == ExpressionType.UnaryFunction:
            return self._detect_syn_convexity(f, g, ctx)
        elif g_type == ExpressionType.Linear and f_type == ExpressionType.UnaryFunction:
            return self._detect_syn_convexity(g, f, ctx)
        return None

    def _detect_syn_convexity_outer(self, linear_expr, sum_expr, ctx):
        non_div_children = []
        for expr in sum_expr.children:
            if expr.expression_type == ExpressionType.Division:
                cvx_num = ctx.convexity(expr.children[0])
                if not (expr.children[1] is linear_expr and cvx_num.is_linear()):
                    return None
            else:
                non_div_children.append(expr)

        if len(non_div_children) != 1:
            return

        # Drill down to find g
        curr_expr = non_div_children[0]
        sign = 1.0
        while True:
            is_unary_func = curr_expr.expression_type == ExpressionType.UnaryFunction
            if is_unary_func and curr_expr.func_type in self._convex_concave_functions:
                break  # unary function found
            elif curr_expr.expression_type == ExpressionType.Product:
                a, b = curr_expr.children
                a_type = a.expression_type
                b_type = b.expression_type
                if a_type == ExpressionType.Constant and b_type == ExpressionType.UnaryFunction:
                    curr_expr = b
                    sign *= np.sign(a.value)
                elif b_type == ExpressionType.Constant and a_type == ExpressionType.UnaryFunction:
                    curr_expr = a
                    sign *= np.sign(b.value)
                else:
                    return
            elif curr_expr.expression_type == ExpressionType.Negation:
                curr_expr = curr_expr.children[0]
                sign *= -1.0
            else:
                return

        cvx = self._detect_syn_convexity(linear_expr, curr_expr, ctx)

        if cvx is None:
            return None
        if sign > 0:
            return cvx
        return cvx.negate()

    def _detect_syn_convexity(self, linear_expr, unary_expr, _ctx):
        g_func = unary_expr.func_type
        denominator = self._denominator(unary_expr.children[0])

        if denominator is None:
            return None

        if denominator is linear_expr:
            if g_func in self._convex_functions:
                return Convexity.Convex
            elif g_func in self._concave_functions:
                return Convexity.Concave
            else:
                raise RuntimeError('unreachable')
        return None

    def _denominator(self, expr):
        expr_type = expr.expression_type
        if expr_type == ExpressionType.Division:
            return expr.children[1]
        elif expr_type == ExpressionType.Sum:
            divs = []
            for child in expr.children:
                den = self._denominator(child)
                if den is not None:
                    divs.append(den)
            if len(divs) == 1:
                return divs[0]
        return None



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name
"""Arbitrary precision mathematical constants and comparison."""
import mpmath


mpf = mpmath.mpf
make_number = mpmath.mpf
inf = make_number('inf')
zero = make_number(0)
pi = mpmath.pi
isnan = mpmath.isnan
isinf = mpmath.isinf


def _declare_unary_function(name, fun):
    # skip rounding mode since we don't round
    globals()[name] = lambda n, _: fun(n)


def _declare_binary_function(name, fun):
    # skip rounding mode since we don't round
    globals()[name] = lambda x, y, _: fun(x, y)


_UNARY_FUNCTIONS = ['sqrt', 'log', 'exp', 'sin', 'asin', 'cos', 'acos', 'tan', 'atan']
for fun in _UNARY_FUNCTIONS:
    _declare_unary_function(fun, getattr(mpmath, fun))

_BINARY_FUNCTIONS = ['power']
for fun in _BINARY_FUNCTIONS:
    _declare_binary_function(fun, getattr(mpmath, fun))


def down(f):
    """Perform computation rounding down."""
    return f()


def up(f):
    """Perform computation rounding down."""
    return f()


def min_(*args):
    """Returns the minimum."""
    return min(a for a in args if not isnan(a))


def max_(*args):
    """Return the maximum."""
    return max(a for a in args if not isnan(a))


def almosteq(a, b, rel_eps=None, abs_eps=None):
    """Floating point equality check between `a` and `b`."""
    # in mpmath inf != inf, but we want inf == inf
    if abs(a) == inf and abs(b) == inf:
        return (a > 0 and b > 0) or (a < 0 and b < 0)
    return mpmath.almosteq(a, b, rel_eps=rel_eps, abs_eps=abs_eps)


def almostgte(a, b):
    """Return True if a >= b."""
    return a > b or almosteq(a, b)


def almostlte(a, b):
    """Return True if a <= b."""
    return a < b or almosteq(a, b)



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convexity detection rules for division expressions."""
from suspect.convexity.convexity import Convexity
from suspect.convexity.rules.rule import ConvexityRule
from suspect.monotonicity.monotonicity import Monotonicity
from suspect.interval import Interval


class DivisionRule(ConvexityRule):
    """Return convexity of division."""
    def apply(self, expr, convexity, monotonicity, bounds):
        f, g = expr.args
        cvx_f = convexity[f]
        cvx_g = convexity[g]

        mono_f = monotonicity[f]
        mono_g = monotonicity[g]

        bounds_f = bounds[f]
        bounds_g = bounds[g]

        return _division_convexity(cvx_f, cvx_g, mono_f, mono_g, bounds_f, bounds_g)


class ReciprocalRule(ConvexityRule):
    """Return convexity of reciprocal."""
    def apply(self, expr, convexity, monotonicity, bounds):
        g = expr.args[0]
        cvx_f = Convexity.Linear
        cvx_g = convexity[g]

        mono_f = Monotonicity.Constant
        mono_g = monotonicity[g]

        bounds_f = Interval(1.0, 1.0)
        bounds_g = bounds[g]

        return _division_convexity(cvx_f, cvx_g, mono_f, mono_g, bounds_f, bounds_g)


def _division_convexity(cvx_f, cvx_g, mono_f, mono_g, bounds_f, bounds_g):
    if mono_g.is_constant():
        return _division_convexity_constant_g(mono_f, mono_g, bounds_f, bounds_g, cvx_f)
    elif mono_f.is_constant():
        return _division_convexity_constant_f(mono_f, mono_g, bounds_f, bounds_g, cvx_g)
    return Convexity.Unknown


def _division_convexity_constant_g(_mono_f, _mono_g, _bounds_f, bounds_g, cvx_f):
    if cvx_f.is_convex() and bounds_g.is_positive():
        return Convexity.Convex
    elif cvx_f.is_concave() and bounds_g.is_negative():
        return Convexity.Convex
    elif cvx_f.is_concave() and bounds_g.is_positive():
        return Convexity.Concave
    elif cvx_f.is_convex() and bounds_g.is_negative():
        return Convexity.Concave
    return Convexity.Unknown


def _division_convexity_constant_f(_mono_f, _mono_g, bounds_f, bounds_g, cvx_g):
    # want to avoid g == 0
    if 0 in bounds_g:
        return Convexity.Unknown

    if bounds_f.is_nonnegative():
        if bounds_g.is_positive() and cvx_g.is_concave():
            return Convexity.Convex
        elif bounds_g.is_negative() and cvx_g.is_convex():
            return Convexity.Concave
    elif bounds_f.is_nonpositive():
        if bounds_g.is_negative() and cvx_g.is_convex():
            return Convexity.Convex
        elif bounds_g.is_positive() and cvx_g.is_concave():
            return Convexity.Concave
    return Convexity.Unknown



# Copyright 2019 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ConvexityRule(object):
    """Base class for rules that compute convexity properties of an expression."""
    def apply(self, expr, convexity, monotonicity, bounds):
        """Apply rule to `expr`.

        Parameters
        ----------
        expr : Expression
            the expression
        convexity : dict-like
            contains convexity of children
        monotonicity : dict-like
            contains monotonicity of children
        bounds : dict-like
            contains bounds of children
        """
        raise NotImplementedError('apply')



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Monotonicity detection rules for linear and sum expressions."""
from suspect.monotonicity.monotonicity import Monotonicity
from suspect.monotonicity.rules.rule import MonotonicityRule


class LinearRule(MonotonicityRule):
    """Return monotonicity of linear expression."""
    def apply(self, expr, mono, bounds):
        monos = [
            _adjust_monotonicity(mono[child], expr.coefficient(child))
            for child in expr.args
        ]
        return _combine_monotonicities(monos)


class SumRule(MonotonicityRule):
    """Return monotonicity of sum expression."""
    def apply(self, expr, monotonicity, bounds):
        monos = [monotonicity[child] for child in expr.args]
        return _combine_monotonicities(monos)


def _adjust_monotonicity(mono, coef):
    if mono.is_unknown() or mono.is_constant():
        return mono

    if coef > 0:
        return mono
    elif coef < 0:
        return mono.negate()

    # if coef == 0, it's a constant with value 0.0
    return Monotonicity.Constant


def _combine_monotonicities(monos):
    all_const = all([m.is_constant() for m in monos])
    all_nondec = all([m.is_nondecreasing() for m in monos])
    all_noninc = all([m.is_nonincreasing() for m in monos])

    if all_const:
        return Monotonicity.Constant
    elif all_nondec:
        return Monotonicity.Nondecreasing
    elif all_noninc:
        return Monotonicity.Nonincreasing
    return Monotonicity.Unknown



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Monotonicity detection rules for trigonometric expressions."""
from suspect.monotonicity.monotonicity import Monotonicity
from suspect.monotonicity.rules.rule import MonotonicityRule


class SinRule(MonotonicityRule):
    """Return monotonicity of sin."""
    def apply(self, expr, monotonicity, bounds):
        child = expr.args[0]
        child_bounds = bounds[child]
        cos_bounds = child_bounds.cos()
        mono = monotonicity[child]

        if mono.is_nondecreasing() and cos_bounds.is_nonnegative():
            return Monotonicity.Nondecreasing
        elif mono.is_nonincreasing() and cos_bounds.is_nonpositive():
            return Monotonicity.Nondecreasing
        elif mono.is_nonincreasing() and cos_bounds.is_nonnegative():
            return Monotonicity.Nonincreasing
        elif mono.is_nondecreasing() and cos_bounds.is_nonpositive():
            return Monotonicity.Nonincreasing
        return Monotonicity.Unknown # pragma: no cover


class CosRule(MonotonicityRule):
    """Return monotonicity of cos."""
    def apply(self, expr, monotonicity, bounds):
        child = expr.args[0]
        child_bounds = bounds[child]
        sin_bounds = child_bounds.sin()
        mono = monotonicity[child]

        if mono.is_nonincreasing() and sin_bounds.is_nonnegative():
            return Monotonicity.Nondecreasing
        elif mono.is_nondecreasing() and sin_bounds.is_nonpositive():
            return Monotonicity.Nondecreasing
        elif mono.is_nondecreasing() and sin_bounds.is_nonnegative():
            return Monotonicity.Nonincreasing
        elif mono.is_nonincreasing() and sin_bounds.is_nonpositive():
            return Monotonicity.Nonincreasing
        return Monotonicity.Unknown



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from .summary import detect_special_structure
from .pyomo import create_connected_model
from .fbbt import perform_fbbt
from .propagation import propagate_special_structure
from .polynomial_degree import polynomial_degree


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)
logger.addHandler(logging.StreamHandler())



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
import abc
import sys
from suspect.math import inf # pylint: disable=no-name-in-module


class Domain(Enum):
    """The variable domain"""
    REALS = 0
    INTEGERS = 1
    BINARY = 2


class Sense(Enum):
    """The objective function sense"""
    MINIMIZE = 0
    MAXIMIZE = 1


class Expression(metaclass=abc.ABCMeta):
    """The base class for all expressions objects in the DAG"""
    is_source = False
    is_sink = False

    def __init__(self, children=None):
        if children is None:
            children = []

        self._children = children

    @property
    def args(self):
        return self._children

    @property
    def args(self):
        return self._children

    def is_variable_type(self):
        return False

    def is_constant(self):
        return False


class Objective(Expression):
    is_sink = True

    def __init__(self, name, sense=None, children=None):
        super().__init__(children)
        if sense is None:
            sense = Sense.MINIMIZE
        self.sense = sense
        self.name = name

    @property
    def body(self):
        return self.args[0]

    def is_expression_type(self):
        return True

    def is_minimizing(self):
        return self.sense == Sense.MINIMIZE

    def is_maximizing(self):
        return self.sense == Sense.MAXIMIZE


class BoundedExpression(Expression):
    def __init__(self, lower_bound, upper_bound, children=None):
        super().__init__(children)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def bounded_below(self):
        return self.lower_bound is not None and self.lower_bound != -inf

    def bounded_above(self):
        return self.upper_bound is not None and self.upper_bound != inf


class Constraint(BoundedExpression):
    is_sink = True

    def __init__(self, name, lower_bound, upper_bound, children=None):
        super().__init__(lower_bound, upper_bound, children)
        self.name = name

    @property
    def body(self):
        return self.args[0]

    def is_expression_type(self):
        return True

    def is_equality(self):
        return self.lower_bound == self.upper_bound

    def __str__(self):
        return 'Constraint(name={}, lower_bound={}, upper_bound={}, args={})'.format(
            self.name, self.lower_bound, self.upper_bound, self.args
        )


class Variable(BoundedExpression):
    is_source = True

    def __init__(self, name, lower_bound, upper_bound, domain=None):
        super().__init__(lower_bound, upper_bound, None)
        self.domain = domain
        self.name = name

    def is_binary(self):
        return self.domain == Domain.BINARY

    def is_integer(self):
        return self.domain == Domain.INTEGERS

    def is_real(self):
        return self.domain == Domain.REALS

    def is_constant(self):
        # TODO(fracek): if we ever support fixing variables, change this
        return False

    def __str__(self):
        return 'Variable(name={}, lower_bound={}, upper_bound={}, domain={})'.format(
            self.name, self.lower_bound, self.upper_bound, self.domain
        )


class Constant(BoundedExpression):
    is_source = True

    def __init__(self, value):
        super().__init__(value, value, None)

    @property
    def value(self):
        assert self.lower_bound == self.upper_bound
        return self.lower_bound

    def is_constant(self):
        return True



#  Copyright 2020 Francesco Ceccon
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pyomo.environ as pyo


VAR_TYPES = [pyo.Reals, pyo.Integers, pyo.Binary]
DEFAULT_VAR_TYPE = {
    'C': 0,
    'I': 1,
    'B': 2,
}


def _is_comment(line):
    if len(line) == 0:
        return True
    return line[0] in ['!', '#', '%']


def _split_line(line, count, parse):
    if count == 1:
        return parse(line.split()[0])
    elements = line.split()[:count]
    return [p(v) for p, v in zip(parse, elements)]


class _QPLibParser:
    def __init__(self):
        self._state = 'name'

        self._name = None
        self._type = None
        self._sense = None

        self._num_variables = None

        # Only if <> **N, **B
        self._num_constraints = None

        # Only if <> L**
        self._obj_q_nnz_count = None
        self._obj_q_nnz_iter = None
        self._obj_q_entries = dict()

        self._b0_default = None

        self._obj_b_nnz_count = None
        self._obj_b_nnz_iter = None
        self._obj_b_entries = dict()

        self._obj_const = None

        # Only if <> **N, **B, **L
        self._con_q_nnz_count = None
        self._con_q_nnz_iter = None
        self._con_q_entries = dict()

        # Only if <> **N, **B
        self._con_b_nnz_count = None
        self._con_b_nnz_iter = None
        self._con_b_entries = dict()

        self._infinity = None

        # Only if <> **N, **B
        self._cl_default = None
        self._cl_nnz_count = None
        self._cl_nnz_iter = None
        self._cl_entries = dict()

        # Only if <> **N, **B
        self._cu_default = None
        self._cu_nnz_count = None
        self._cu_nnz_iter = None
        self._cu_entries = dict()

        # Only if <> *B*
        self._l_default = None
        self._l_nnz_count = None
        self._l_nnz_iter = None
        self._l_entries = dict()

        # Only if <> *B*
        self._u_default = None
        self._u_nnz_count = None
        self._u_nnz_iter = None
        self._u_entries = dict()

        # Only if <> *C*, *B*, *I*
        self._v_default = None
        self._v_nnz_count = None
        self._v_nnz_iter = None
        self._v_entries = dict()

        self._x0_default = None
        self._x0_nnz_count = None
        self._x0_nnz_iter = None
        self._x0_entries = dict()

        # Only if <> **N, **B
        self._y0_default = None
        self._y0_nnz_count = None
        self._y0_nnz_iter = None
        self._y0_entries = dict()

        self._z0_default = None
        self._z0_nnz_count = None
        self._z0_nnz_iter = None
        self._z0_entries = dict()

        self._var_name_nnz_count = None
        self._var_name_nnz_iter = None
        self._var_name_nnz_entries = dict()

        self._con_name_nnz_count = None
        self._con_name_nnz_iter = None
        self._con_name_nnz_entries = dict()

    def process_line(self, line):
        line = line.strip()
        if _is_comment(line):
            return

        processed = False
        while not processed:
            callback = getattr(self, '_process_{}'.format(self._state))
            processed = callback(line)

    def _process_name(self, line):
        self._name = _split_line(line, 1, str)
        self._state = 'type'
        return True

    def _process_type(self, line):
        self._type = _split_line(line, 1, str)
        self._state = 'sense'
        return True

    def _process_sense(self, line):
        line = _split_line(line, 1, str)
        if line.lower() == 'minimize':
            self._sense = pyo.minimize
        else:
            self._sense = pyo.maximize
        self._state = 'num_variables'
        return True

    def _process_num_variables(self, line):
        self._num_variables = _split_line(line, 1, int)
        self._state = 'num_constraints'
        return True

    def _process_num_constraints(self, line):
        if self._type[2] not in ['B', 'N']:
            self._num_constraints = _split_line(line, 1, int)
            self._state = 'obj_q_nnz_count'
            return True
        else:
            self._num_constraints = 0
            self._state = 'obj_q_nnz_count'
            return False

    def _process_obj_q_nnz_count(self, line):
        if self._type[0] in ['L']:
            self._obj_q_nnz_count = 0
            self._state = 'b_default'
            return False
        else:
            self._obj_q_nnz_count = _split_line(line, 1, int)
            self._obj_q_nnz_iter = 0
            self._state = 'obj_q_entries'
            return True

    def _process_obj_q_entries(self, line):
        if self._obj_q_nnz_iter >= self._obj_q_nnz_count:
            self._state = 'b_default'
            return False
        else:
            self._obj_q_nnz_iter += 1
            i, j, q = _split_line(line, 3, [int, int, float])
            self._obj_q_entries[(i, j)] = q
            return True

    def _process_b_default(self, line):
        self._b0_default = _split_line(line, 1, float)
        self._state = 'obj_b_nnz_count'
        return True

    def _process_obj_b_nnz_count(self, line):
        self._obj_b_nnz_count = _split_line(line, 1, int)
        self._obj_b_nnz_iter = 0
        self._state = 'obj_b_entries'
        return True

    def _process_obj_b_entries(self, line):
        if self._obj_b_nnz_iter >= self._obj_b_nnz_count:
            self._state = 'obj_const'
            return False
        else:
            self._obj_b_nnz_iter += 1
            j, b = _split_line(line, 2, [int, float])
            self._obj_b_entries[j] = b
            return True

    def _process_obj_const(self, line):
        self._obj_const = _split_line(line, 1, float)
        self._state = 'con_q_nnz_count'
        return True

    def _process_con_q_nnz_count(self, line):
        if self._type[2] in ['N', 'B', 'L']:
            self._state = 'con_b_nnz_count'
            return False
        else:
            self._con_q_nnz_count = _split_line(line, 1, int)
            self._con_q_nnz_iter = 0
            self._state = 'con_q_entries'
            return True

    def _process_con_q_entries(self, line):
        if self._con_q_nnz_iter >= self._con_q_nnz_count:
            self._state = 'con_b_nnz_count'
            return False
        else:
            self._con_q_nnz_iter += 1
            i, h, k, q = _split_line(line, 4, [int, int, int, float])
            if i not in self._con_q_entries:
                self._con_q_entries[i] = dict()
            self._con_q_entries[i][(h, k)] = q
            return True

    def _process_con_b_nnz_count(self, line):
        if self._type[2] in ['N', 'B']:
            self._state = 'infinity'
            return False
        else:
            self._con_b_nnz_count = _split_line(line, 1, int)
            self._con_b_nnz_iter = 0
            self._state = 'con_b_entries'
            return True

    def _process_con_b_entries(self, line):
        if self._con_b_nnz_iter >= self._con_b_nnz_count:
            self._state = 'infinity'
            return False
        else:
            self._con_b_nnz_iter += 1
            i, j, b = _split_line(line, 3, [int, int, float])
            if i not in self._con_b_entries:
                self._con_b_entries[i] = dict()
            self._con_b_entries[i][j] = b
            return True

    def _process_infinity(self, line):
        self._infinity = _split_line(line, 1, float)
        self._state = 'cl_default'
        return True

    def _process_cl_default(self, line):
        if self._type[2] in ['B', 'N']:
            self._state = 'l_default'
            return False
        else:
            self._cl_default = _split_line(line, 1, float)
            self._state = 'cl_nnz_count'
            return True

    def _process_cl_nnz_count(self, line):
        self._cl_nnz_count = _split_line(line, 1, int)
        self._cl_nnz_iter = 0
        self._state = 'cl_entries'
        return True

    def _process_cl_entries(self, line):
        if self._cl_nnz_iter >= self._cl_nnz_count:
            self._state = 'cu_default'
            return False
        else:
            self._cl_nnz_iter += 1
            i, c = _split_line(line, 2, [int, float])
            self._cl_entries[i] = c
            return True

    def _process_cu_default(self, line):
        self._cu_default = _split_line(line, 1, float)
        self._state = 'cu_nnz_count'
        return True

    def _process_cu_nnz_count(self, line):
        self._cu_nnz_count = _split_line(line, 1, int)
        self._cu_nnz_iter = 0
        self._state = 'cu_entries'
        return True

    def _process_cu_entries(self, line):
        if self._cu_nnz_iter >= self._cu_nnz_count:
            self._state = 'l_default'
            return False
        else:
            self._cu_nnz_iter += 1
            i, c = _split_line(line, 2, [int, float])
            self._cu_entries[i] = c
            return True

    def _process_l_default(self, line):
        if self._type[1] in ['B']:
            self._state = 'u_default'
            self._l_default = 0
            return False
        else:
            self._l_default = _split_line(line, 1, float)
            self._state = 'l_nnz_count'
            return True

    def _process_l_nnz_count(self, line):
        self._l_nnz_count = _split_line(line, 1, int)
        self._l_nnz_iter = 0
        self._state = 'l_entries'
        return True

    def _process_l_entries(self, line):
        if self._l_nnz_iter >= self._l_nnz_count:
            self._state = 'u_default'
            return False
        else:
            self._l_nnz_iter += 1
            i, c = _split_line(line, 2, [int, float])
            self._l_entries[i] = c
            return True

    def _process_u_default(self, line):
        if self._type[1] in ['B']:
            self._state = 'v_default'
            self._u_default = 1
            return False
        else:
            self._u_default = _split_line(line, 1, float)
            self._state = 'u_nnz_count'
            return True

    def _process_u_nnz_count(self, line):
        self._u_nnz_count = _split_line(line, 1, int)
        self._u_nnz_iter = 0
        self._state = 'u_entries'
        return True

    def _process_u_entries(self, line):
        if self._u_nnz_iter >= self._u_nnz_count:
            self._state = 'v_default'
            return False
        else:
            self._u_nnz_iter += 1
            i, c = _split_line(line, 2, [int, float])
            self._u_entries[i] = c
            return True

    def _process_v_default(self, line):
        if self._type[1] in ['C', 'B', 'I']:
            self._state = 'x0_default'
            self._v_default = DEFAULT_VAR_TYPE[self._type[1]]
            return False
        self._v_default = _split_line(line, 1, int)
        self._state = 'v_nnz_count'
        return True

    def _process_v_nnz_count(self, line):
        if self._type[1] in ['C', 'B', 'I']:
            self._state = 'x0_nnz_count'
            return False
        else:
            self._v_nnz_count = _split_line(line, 1, int)
            self._v_nnz_iter = 0
            self._state = 'v_entries'
            return True

    def _process_v_entries(self, line):
        if self._v_nnz_iter >= self._v_nnz_count:
            self._state = 'x0_default'
            return False
        else:
            self._v_nnz_iter += 1
            i, c = _split_line(line, 2, [int, int])
            self._v_entries[i] = c
            return True

    def _process_x0_default(self, line):
        self._x0_default = _split_line(line, 1, float)
        self._state = 'x0_nnz_count'
        return True

    def _process_x0_nnz_count(self, line):
        self._x0_nnz_count = _split_line(line, 1, int)
        self._x0_nnz_iter = 0
        self._state = 'x0_entries'
        return True

    def _process_x0_entries(self, line):
        if self._x0_nnz_iter >= self._x0_nnz_count:
            self._state = 'y0_default'
            return False
        else:
            self._x0_nnz_iter += 1
            i, c = _split_line(line, 2, [int, float])
            self._x0_entries[i] = c
            return True

    def _process_y0_default(self, line):
        if self._type[2] in ['B', 'N']:
            self._state = 'z0_default'
            return False
        else:
            self._y0_default = _split_line(line, 1, float)
            self._state = 'y0_nnz_count'
            return True

    def _process_y0_nnz_count(self, line):
        self._y0_nnz_count = _split_line(line, 1, int)
        self._y0_nnz_iter = 0
        self._state = 'y0_entries'
        return True

    def _process_y0_entries(self, line):
        if self._y0_nnz_iter >= self._y0_nnz_count:
            self._state = 'z0_default'
            return False
        else:
            self._y0_nnz_iter += 1
            i, c = _split_line(line, 2, [int, float])
            self._y0_entries[i] = c
            return True

    def _process_z0_default(self, line):
        self._z0_default = _split_line(line, 1, float)
        self._state = 'z0_nnz_count'
        return True

    def _process_z0_nnz_count(self, line):
        self._z0_nnz_count = _split_line(line, 1, int)
        self._z0_nnz_iter = 0
        self._state = 'z0_entries'
        return True

    def _process_z0_entries(self, line):
        if self._z0_nnz_iter >= self._y0_nnz_count:
            self._state = 'var_name_nnz_count'
            return False
        else:
            self._z0_nnz_iter += 1
            i, c = _split_line(line, 2, [int, float])
            self._z0_entries[i] = c
            return True

    def _process_var_name_nnz_count(self, line):
        self._var_name_nnz_count = _split_line(line, 1, int)
        self._var_name_nnz_iter = 0
        self._state = 'var_name_entries'
        return True

    def _process_var_name_entries(self, line):
        if self._var_name_nnz_iter >= self._var_name_nnz_count:
            self._state = 'con_name_nnz_count'
            return False
        else:
            self._var_name_nnz_iter += 1
            i, n = _split_line(line, 2, [int, str])
            self._var_name_nnz_entries[i] = n
            return True

    def _process_con_name_nnz_count(self, line):
        self._con_name_nnz_count = _split_line(line, 1, int)
        self._con_name_nnz_iter = 0
        self._state = 'con_name_entries'
        return True

    def _process_con_name_entries(self, line):
        if self._con_name_nnz_iter >= self._con_name_nnz_count:
            self._state = 'done'
            return False
        else:
            self._con_name_nnz_iter += 1
            i, n = _split_line(line, 2, [int, str])
            self._con_name_nnz_entries[i] = n
            return True

    def _process_done(self, line):
        raise RuntimeError('Finished parsing file but read line: {}'.format(line))

    @property
    def model(self):
        model = pyo.ConcreteModel(name=self._name)

        # build variables
        default_var_lb = self._l_default
        default_var_ub = self._u_default
        default_var_x0 = self._x0_default
        variables = []
        for i in range(self._num_variables):
            idx = i + 1
            var_type = VAR_TYPES[self._v_entries.get(idx, self._v_default)]
            var_name = self._var_name_nnz_entries.get(idx, 'x{}'.format(idx))
            var_lb = self._l_entries.get(idx, default_var_lb)
            var_ub = self._u_entries.get(idx, default_var_ub)
            var_x0 = self._x0_entries.get(idx, default_var_x0)
            if var_type in [pyo.Integers, pyo.Binary]:
                var_x0 = int(var_x0)
            var = pyo.Var(
                name=var_name,
                domain=var_type,
                bounds=(var_lb, var_ub),
                initialize=lambda _: var_x0
            )
            variables.append(var)
            setattr(model, var_name, var)

        # build objective
        obj_quad = self._obj_const
        for (i, j), q in self._obj_q_entries.items():
            # d = 2.0 if i == j else 1.0
            d = 2.0
            obj_quad += (q / d) * variables[i-1] * variables[j-1]

        obj_linear = pyo.quicksum(
            [v * self._obj_b_entries.get(i + 1, self._b0_default) for i, v in enumerate(variables)],
            linear=False
        )
        model.obj = pyo.Objective(expr=obj_quad + obj_linear, sense=self._sense)

        # build constraints
        for i in range(self._num_constraints):
            con_idx = i + 1
            con_quad = 0.0
            if con_idx not in self._con_q_entries and con_idx not in self._con_b_entries:
                continue
            if con_idx in self._con_q_entries:
                for (h, k), q in self._con_q_entries[con_idx].items():
                    # d = 2.0 if h == k else 1.0
                    d = 2.0
                    con_quad += (q / d) * variables[h-1] * variables[k-1]
            con_linear = 0.0
            if con_idx in self._con_b_entries:
                con_linear = pyo.quicksum(
                    [b * variables[j-1] for j, b in self._con_b_entries[con_idx].items()],
                    linear=False,
                )
            con_name = self._con_name_nnz_entries.get(i+1, 'e{}'.format(i+1))
            con_lb = self._cl_entries.get(i+1, self._cl_default)
            con_ub = self._cu_entries.get(i+1, self._cu_default)
            con_expr = pyo.inequality(con_lb, con_quad + con_linear, con_ub)
            con = pyo.Constraint(name=con_name, expr=con_expr)
            setattr(model, con_name, con)

        return model


def read_qplib(filename, **kwargs):
    parser = _QPLibParser()
    with open(filename) as f:
        for line in f:
            parser.process_line(line)
    return parser.model


# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convexity detection rules for convex functions."""
from suspect.convexity.convexity import Convexity
from suspect.convexity.rules.rule import ConvexityRule


class ConvexFunctionRule(ConvexityRule):
    """Return convexity of convex function."""
    def apply(self, expr, convexity, _mono, _bounds):
        child = expr.args[0]
        cvx = convexity[child]
        if cvx.is_convex():
            return Convexity.Convex
        return Convexity.Unknown


class ExpRule(ConvexFunctionRule):
    """Return convexity of exp."""
    pass



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import pyomo.environ as aml
from suspect.pyomo.expr_dict import *
from suspect.float_hash import RoundFloatHasher
from tests.fixtures import model


def test_expr_hash_linear_expr(model):
    model.cons1 = aml.Constraint(
        rule=lambda m: sum(m.x[i] for i in m.I) == 2
    )
    model.cons2 = aml.Constraint(
        rule=lambda m: sum(m.x[i] for i in m.I) == 2
    )
    model.cons3 = aml.Constraint(
        rule=lambda m: sum(m.x[i] for i in m.I[:2]) == 3
    )

    hash_cons1 = expr_hash(model.cons1.expr)
    hash_cons2 = expr_hash(model.cons2.expr)
    hash_cons3 = expr_hash(model.cons3.expr)

    assert hash_cons1 == hash_cons2
    # TODO: fix too common collisions
    # assert hash_cons1 != hash_cons3

    # we want hashing of linear expr be associative
    model.cons4 = aml.Constraint(
        rule=lambda m: sum(m.x[i] for i in m.I[::-1]) == 2
    )

    hash_cons4 = expr_hash(model.cons4.expr)
    assert hash_cons1 == hash_cons4


def test_expr_hash_product_expr(model):
    model.cons1 = aml.Constraint(
        model.I,
        rule=lambda m, i: 0 <= m.x[i] * m.y[i, 0] <= 10
    )

    hash_cons1 = expr_hash(model.cons1[0].expr)
    for i in model.I:
        for j in model.J:
            if i == 0 and j == 0:
                continue
            # Collisions could happen, so a failure here is expected
            # in some rare cases
            ex = -10 <= model.x[i] * model.y[i, j] <= 100
            assert hash_cons1 != expr_hash(ex)


def test_expr_hash_unary_expr(model):
    model.cons1 = aml.Constraint(
        model.I,
        rule=lambda m, i: aml.cos(m.x[i]) <= 0.5
    )

    hash_cons1 = expr_hash(model.cons1[2].expr)
    ex1 = aml.cos(model.x[2]) <= 0.5
    assert hash_cons1 == expr_hash(ex1)


def test_expr_hash_float_hasher(model):
    model.cons1 = aml.Constraint(
        rule=lambda m: sum(m.x[i] for i in m.I) == 2
    )
    model.cons2 = aml.Constraint(
        rule=lambda m: sum((i+1)*m.x[i] for i in m.I) == 2
    )

    h = RoundFloatHasher(2)

    assert expr_hash(model.cons1.expr, h) != expr_hash(model.cons2.expr, h)
    assert expr_hash(model.cons1.expr) != expr_hash(model.cons2.expr, h)

    model.cons3 = aml.Constraint(
        model.I,
        rule=lambda m, i: m.x[i] == i
    )
    model.cons4 = aml.Constraint(
        model.I,
        rule=lambda m, i: m.x[i] == (i + 1)
    )

    assert expr_hash(model.cons3[0], h) != expr_hash(model.cons4[0], h)


def test_expr_equal(model):
    assert not expr_equal(1.2, 2.3)
    assert expr_equal(model.x[1], model.x[1])
    assert expr_equal(model.x[1] + 1.23, model.x[1] + 1.23)
    assert not expr_equal(model.x[1], model.y[1, 0])
    assert not expr_equal(model.x[1], 2)

    sum1 = sum(model.x[i] for i in model.I)
    sum2 = sum(model.x[i] for i in model.I)
    sum3 = sum((i + 1) * model.x[i] for i in model.I)
    sum4 = sum(i * model.x[i] for i in model.I)
    sum5 = sum((i + 1) * model.x[i] for i in model.I)
    sum6 = sum(model.y[i, 0] for i in model.I)

    assert expr_equal(sum1, sum2)
    assert not expr_equal(sum1, sum3)
    assert not expr_equal(sum1, sum4)
    assert expr_equal(sum3, sum5)
    assert not expr_equal(sum1, sum6)

    fun1 = aml.cos(sum1)
    fun2 = aml.cos(sum2)
    fun3 = aml.sin(sum1)

    assert expr_equal(fun1, fun2)
    assert not expr_equal(fun1, fun3)
    assert expr_equal(fun1 * fun3, fun2 * fun3)
    assert expr_equal(fun1 / fun3, fun2 / fun3)


def test_expr_map(model):
    model.cons1 = aml.Constraint(
        rule=lambda m: sum(m.x[i] for i in m.I) == 2.34
    )
    model.cons2 = aml.Constraint(
        rule=lambda m: sum((i+1)*m.x[i] for i in m.I) == 3
    )

    memo = ExpressionDict()

    assert memo[model.cons1.expr] is None

    memo[model.cons1.expr] = 42
    assert 42 == memo[model.cons1.expr]

    assert memo[model.cons2.expr] is None

    memo[model.cons2.expr] = 123
    assert 123 == memo[model.cons2.expr]

    memo[model.cons1.expr] = 100
    assert 100 == memo[model.cons1.expr]

    assert len(memo) == 2



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Iterators over vertices of ProblemDag."""
import abc
from suspect.pyomo.expressions import nonpyomo_leaf_types
from suspect.interfaces import ForwardIterator, BackwardIterator
from suspect.dag.dag import ProblemDag
from suspect.dag.expressions import Expression
from suspect.dag.vertices_list import VerticesList


class ForwardDagIterator(object):
    def iterate(self, problem, visitor, *args, starting_vertices=None):
        sources = set(id(s) for s in problem._sources)
        if starting_vertices is None:
            starting_vertices = sources
        else:
            starting_vertices = set(id(v) for v in starting_vertices) | sources

        results = {}
        changed = set()
        for vertex in problem.vertices:
            has_changed = visitor.visit(vertex, *args)
            if has_changed:
                changed.add(id(vertex))
        return list(changed)


class _DagIterator(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _get_next_vertices(self, problem, expr):
        pass

    @abc.abstractproperty
    def _reverse(self):
        pass

    def _iterate(self, problem, visitor, starting_vertices, *args):
        changed_vertices = []
        vertices = VerticesList(reverse=self._reverse)
        for vertex in starting_vertices:
            vertices.append(vertex, problem.depth(vertex))
        seen = set()
        while vertices:
            curr_vertex = vertices.pop()
            if id(curr_vertex) in seen:
                continue
            has_changes = visitor.visit(curr_vertex, *args)
            seen.add(id(curr_vertex))

            if has_changes:
                changed_vertices.append(curr_vertex)
                for vertex in self._get_next_vertices(problem, curr_vertex):
                    if id(vertex) not in seen:
                        vertices.append(vertex, problem.depth(vertex))

        return changed_vertices


class DagForwardIterator(_DagIterator, ForwardIterator):
    """Forward iterator over suspect.dag.ProblemDag vertices."""
    def _get_next_vertices(self, problem, expr):
        if type(expr) in nonpyomo_leaf_types:
            return []
        return problem.parents(expr)

    @property
    def _reverse(self):
        return False

    def iterate(self, problem, visitor, *args, **kwargs): # pylint: disable=missing-docstring
        starting_vertices = kwargs.pop('starting_vertices', None)
        # pylint: disable=protected-access
        if starting_vertices is None:
            starting_vertices = problem._sources
        else:
            starting_vertices = problem._sources + starting_vertices

        return super()._iterate(
            problem,
            visitor,
            starting_vertices,
            *args,
        )


class DagBackwardIterator(_DagIterator, BackwardIterator):
    """Backward iterator over suspect.dag.ProblemDag vertices."""
    def _get_next_vertices(self, _problem, expr):
        if not expr.is_expression_type():
            return []
        return expr.args

    @property
    def _reverse(self):
        return True

    def iterate(self, problem, visitor, *args, **kwargs): # pylint: disable=missing-docstring
        starting_vertices = kwargs.pop('starting_vertices', None)
        # pylint: disable=protected-access
        if starting_vertices is None:
            starting_vertices = problem._sinks
        else:
            starting_vertices = problem._sinks + starting_vertices

        return super()._iterate(
            problem,
            visitor,
            starting_vertices,
            *args,
        )



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Visitor applying rules for monotonicity propagation."""
from suspect.interfaces import CombineUnaryFunctionRules
from suspect.monotonicity.rules import *
from suspect.pyomo.expressions import (
    nonpyomo_leaf_types,
    NumericConstant,
    Var,
    Constraint,
    Objective,
    MonomialTermExpression,
    ProductExpression,
    DivisionExpression,
    ReciprocalExpression,
    AbsExpression,
    LinearExpression,
    SumExpression,
    PowExpression,
    NegationExpression,
    UnaryFunctionExpression,
)
from suspect.pyomo.quadratic import QuadraticExpression
from suspect.visitor import Visitor

_expr_to_rule_map = dict()
_expr_to_rule_map[NumericConstant] = ConstantRule()
_expr_to_rule_map[Var] = VariableRule()
_expr_to_rule_map[Constraint] = ConstraintRule()
_expr_to_rule_map[Objective] = ObjectiveRule()
_expr_to_rule_map[MonomialTermExpression] = ProductRule()
_expr_to_rule_map[ProductExpression] = ProductRule()
_expr_to_rule_map[DivisionExpression] = DivisionRule()
_expr_to_rule_map[ReciprocalExpression] = ReciprocalRule()
_expr_to_rule_map[LinearExpression] = LinearRule()
_expr_to_rule_map[SumExpression] = SumRule()
_expr_to_rule_map[PowExpression] = PowerRule()
_expr_to_rule_map[NegationExpression] = NegationRule()
_expr_to_rule_map[AbsExpression] = AbsRule()
_expr_to_rule_map[QuadraticExpression] = QuadraticRule()
_expr_to_rule_map[UnaryFunctionExpression] = CombineUnaryFunctionRules({
    'abs': AbsRule(),
    'sqrt': SqrtRule(),
    'exp': ExpRule(),
    'log': LogRule(),
    'tan': TanRule(),
    'atan': AtanRule(),
    'sin': SinRule(),
    'asin': AsinRule(),
    'cos': CosRule(),
    'acos': AcosRule(),
})


def propagate_expression_monotonicity(expr, mono, bounds):
    if type(expr) in nonpyomo_leaf_types:
        rule = _expr_to_rule_map[NumericConstant]
    elif expr.is_constant():
        rule = _expr_to_rule_map[NumericConstant]
    elif expr.is_variable_type():
        rule = _expr_to_rule_map[Var]
    else:
        assert expr.is_expression_type()
        rule = _expr_to_rule_map[type(expr)]
    return rule.apply(expr, mono, bounds)


class MonotonicityPropagationVisitor(Visitor):
    """Visitor applying monotonicity rules."""
    def handle_result(self, expr, result, monotonicity):
        monotonicity[expr] = result
        return True

    def visit_expression(self, expr, mono, bounds):
        return True, propagate_expression_monotonicity(expr, mono, bounds)



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Number
from functools import reduce
from suspect.math import almosteq
from suspect.float_hash import RoundFloatHasher
from pyomo.core.expr.numeric_expr import (
    nonpyomo_leaf_types,
    NumericConstant,
    LinearExpression,
    UnaryFunctionExpression,
)
from pyomo.core.expr.visitor import ExpressionValueVisitor
from pyomo.core.base import _VarData
from pyomo.core.base.constraint import _GeneralConstraintData


class ExpressionHasherVisitor(ExpressionValueVisitor):
    def __init__(self, float_hasher=None):
        if float_hasher is None:
            float_hasher = RoundFloatHasher()
        self._float_hasher = float_hasher

    def hash(self, expr):
        if expr.__class__ in nonpyomo_leaf_types:
            return self._float_hasher.hash(float(expr))
        return self.memo[id(expr)]

    def set_hash(self, expr, h):
        self.memo[id(expr)] = h

    def _type_xor_args(self, expr, values):
        return hash(type(expr)) ^ reduce(lambda x, y: x ^ y, values, 0)

    def visiting_potential_leaf(self, node):
        if node.__class__ in nonpyomo_leaf_types:
            return True, self._float_hasher.hash(float(node))

        if node.is_constant():
            return True, self._float_hasher.hash(float(node.value))

        if node.is_variable_type():
            h = hash(_VarData) ^ id(node)
            return True, h

        return False, None

    def visit(self, node, values):
        return self._type_xor_args(node, values)

    def visit_linear(self, expr):
        coef = expr._coef
        hashes = [
            self.hash(e) ^ self.hash(coef[id(e)])
            for e in expr._args
        ]
        h = hash(type(expr)) ^ reduce(lambda x, y: x ^ y, hashes, 0)
        self.set_hash(expr, h)


def expr_hash(expr, float_hasher=None):
    if isinstance(expr, _GeneralConstraintData):
        expr = expr.expr

    hasher = ExpressionHasherVisitor(float_hasher=float_hasher)
    return hasher.dfs_postorder_stack(expr)


def _is_leaf_node(expr):
    return isinstance(expr, (Number, NumericConstant, _VarData))


def expr_equal(expr1, expr2):
    stack = [(expr1, expr2)]
    while len(stack) > 0:
        e1, e2 = stack.pop()

        if type(e1) != type(e2):
            return False

        if _is_leaf_node(e1):
            if isinstance(e1, Number):
                if not almosteq(e1, e2):
                    return False

            if isinstance(e1, NumericConstant):
                if not almosteq(e1.value, e2.value):
                    return False

            if isinstance(e1, _VarData):
                if id(e1) != id(e2):
                    return False
        else:
            if len(e1._args_) != len(e2._args_):
                return False

            if isinstance(e1, LinearExpression):
                if len(e1.linear_vars) != len(e2.linear_vars):
                    return False

                for v1, v2 in zip(e1.linear_vars, e2.linear_vars):
                    if id(v1) != id(v2):
                        return False

                for c1, c2 in zip(e2.linear_coefs, e2.linear_coefs):
                    if not almosteq(c1, c2):
                        return False

                if not almosteq(e1.constant, e2.constant):
                    return False

            if isinstance(e1, UnaryFunctionExpression):
                if e1.name != e2.name:
                    return False

            # all checks passed, check args
            for a1, a2 in zip(e1._args_, e2._args_):
                stack.append((a1, a2))

    return True


class ExpressionDict(object):
    def __init__(self, float_hasher=None):
        self._float_hasher = float_hasher
        self._data = {}

    def _hash(self, expr):
        return expr_hash(expr)

    def set(self, expr, v):
        h = self._hash(expr)
        if h not in self._data:
            self._data[h] = [(expr, v)]
        else:
            data = self._data[h]
            for i in range(len(data)):
                old_exp, old_value = data[i]
                if expr_equal(old_exp, expr):
                    data[i] = (expr, v)
                    return
            # no matching expr found, just append
            data.append((expr, v))

    def get(self, expr):
        h = self._hash(expr)
        if h not in self._data:
            return None
        data = self._data[h]
        for (target_expr, value) in data:
            if expr_equal(target_expr, expr):
                return value
        return None

    def __len__(self):
        length = 0
        for _, values in self._data.items():
            length += len(values)
        return length

    def __getitem__(self, expr):
        return self.get(expr)

    def __setitem__(self, expr, value):
        return self.set(expr, value)

    def _dump(self):
        print('Begin ExpressionDict Dump')
        for _, values in self._data.items():
            for expr, v in values:
                print('[{}] -> {}'.format(expr, v))
        print('End')


class TightestExpressionDict(ExpressionDict):
    """Like `ExpressionDict`, but when setting the bounds it will tighten them."""
    def tighten(self, expr, value):
        if value is not None:
            old_bound = self.get(expr)
            if old_bound is None:
                new_bound = value
            else:
                new_bound = old_bound.tighten(value)
            self.set(expr, new_bound)



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: skip-file

import numpy as np
import pyomo.environ as pe
import pytest
from hypothesis import given
import suspect.pyomo.expressions as dex
from suspect.fbbt.tightening.quadratic import *
from suspect.fbbt.tightening.rules import *
from suspect.fbbt.tightening.visitor import BoundsTighteningVisitor
from suspect.interval import Interval as I
from tests.strategies import reals


@pytest.fixture
def visitor():
    return BoundsTighteningVisitor()


@given(reals(), reals())
def test_constraint_rule(visitor, bound1, bound2):
    lower_bound = min(bound1, bound2)
    upper_bound = max(bound1, bound2)
    child = pe.Var()
    constraint = dex.Constraint(
        'test', lower_bound=lower_bound, upper_bound=upper_bound, children=[child]
    )
    matched, result = visitor.visit_expression(constraint, None)
    assert matched
    assert result[0] == I(lower_bound, upper_bound)


class TestSumRule:
    children = None
    expr = None

    def _rule_result(self, visitor, children_bounds, expr_bounds):
        children = [pe.Var() for _ in children_bounds]
        expr = sum(children)

        bounds = ComponentMap()
        bounds[expr] = expr_bounds
        for child, child_bounds in zip(children, children_bounds):
            bounds[child] = child_bounds

        self.children = children
        self.expr = expr
        matched, result = visitor.visit_expression(expr, bounds)
        assert matched
        return result

    def test_unbounded_expr(self, visitor):
        new_bounds = self._rule_result(visitor, [I(0, 1), I(-1, 0)], I(0, None))
        assert new_bounds[0] == I(0, None)
        assert new_bounds[1] == I(-1, None)

    def test_bounded_expr(self, visitor):
        expr_bounds = I(-5, 5)
        children_bounds = [I(1, 2), I(-2, 2), I(-10, 10)]
        new_bounds = self._rule_result(visitor, children_bounds, expr_bounds)
        assert new_bounds[2] == I(-9, 6)


class TestLinearRule:
    children = None
    expr = None

    def _rule_result(self, coefficients, children_bounds, expr_bounds):
        children = [pe.Var() for _ in children_bounds]
        expr = pe.quicksum(
            [child * coef for child, coef in zip(children, coefficients)],
            linear=True
        )

        bounds = ComponentMap()
        bounds[expr] = expr_bounds
        for child, child_bounds in zip(children, children_bounds):
            bounds[child] = child_bounds

        rule = LinearRule()
        self.children = children
        self.expr = expr
        return rule.apply(expr, bounds)

    def test_unbounded_expr(self):
        new_bounds = self._rule_result([-1, 1], [I(0, 1), I(-1, 0)], I(0, None))
        assert new_bounds[0] == I(None, 0)
        assert new_bounds[1] == I(0, None)

    def test_bounded_expr(self):
        expr_bounds = I(-5 , 5)
        children_bounds = [I(-2, -1), I(-2, 2), I(-10, 10)]
        new_bounds = self._rule_result([-1, 1, 2], children_bounds, expr_bounds)
        assert new_bounds[2] == I(-4.5, 3)


class _BilinearTerm:
    def __init__(self, var1, var2, coef):
        self.var1 = var1
        self.var2 = var2
        self.coefficient = coef


class _QuadraticExpression:
    expression_type = ExpressionType.Quadratic

    def __init__(self, children, terms):
        self.args = children
        self.terms = terms


class TestQuadraticRule:
    children = None
    expr = None

    def _rule_result(self, terms, children, children_bounds, expr_bounds):
        expr = _QuadraticExpression(children, terms)

        bounds = ComponentMap()
        bounds[expr] = expr_bounds
        for child, child_bounds in zip(children, children_bounds):
            bounds[child] = child_bounds

        rule = QuadraticRule()
        self.children = children
        self.expr = expr
        return rule.apply(expr, bounds)

    def test_sum_of_squares(self):
        x = pe.Var()
        y = pe.Var()
        new_bounds = self._rule_result(
            [_BilinearTerm(x, x, 1.0), _BilinearTerm(y, y, 2.0)],
            [x, y],
            [I(None, None), I(None, None)],
            I(None, 10),
        )
        assert new_bounds[x] == I(-np.sqrt(10.0), np.sqrt(10))
        assert new_bounds[y] == I(-np.sqrt(5), np.sqrt(5))


class TestPowerRuleConstantExpo:
    base = None

    def _rule_result(self, visitor, expo, expr_bounds):
        base = pe.Var()
        expr = base ** expo
        self.base = base

        bounds = ComponentMap()
        bounds[expr] = expr_bounds
        return visitor.visit_expression(expr, bounds)

    def test_square(self, visitor):
        matched, new_bounds = self._rule_result(visitor, 2.0, I(0, 4))
        assert matched
        assert new_bounds[0] == I(-2, 2)

    def test_non_square(self, visitor):
        matched, new_bounds = self._rule_result(visitor, 3.0, I(0, 4))
        assert not matched
        assert new_bounds is None


class TestUnboundedFunctionsRule:
    @pytest.mark.parametrize('expr_type', [
        pe.sin, pe.cos, pe.tan, pe.asin, pe.acos, pe.atan,
    ])
    @given(a=reals(), b=reals())
    def test_unbounded(self, visitor, a, b, expr_type):
        lower = min(a, b)
        upper = max(a, b)
        child = pe.Var(bounds=(lower, upper))
        expr = expr_type(child)
        bounds = ComponentMap()
        bounds[expr] = I(lower, upper)
        matched, result = visitor.visit_expression(expr, bounds)
        assert matched
        assert result[0] == I(None, None)


class TestVisitor:
    def test_handle_result_with_new_value(self, visitor):
        bounds = ComponentMap()
        expr = pe.Var()
        assert visitor.handle_result(expr, Interval(0, 1), bounds)
        assert bounds[expr] == Interval(0, 1)

    def test_handle_result_with_no_new_value(self, visitor):
        bounds = ComponentMap()
        expr = pe.Var()
        visitor.handle_result(expr, Interval(0, 1), bounds)
        assert not visitor.handle_result(expr, None, bounds)
        assert bounds[expr] == Interval(0, 1)


class _SumExpression:
    expression_type = ExpressionType.Sum

    def __init__(self, children):
        self.children = children

    def nargs(self):
        return len(self.children)


class _LinearExpression:
    expression_type = ExpressionType.Linear

    def __init__(self, coefs, vars):
        self.linear_coefs = coefs
        self.linear_vars = vars
        self.children = vars


class TestUnivariateQuadraticRule:
    def test_result_with_squares(self):
        x = pe.Var()
        y = pe.Var()
        z = pe.Var()
        quadratic = _QuadraticExpression(
            [x, y, z],
            [_BilinearTerm(x, x, 1.0),
             _BilinearTerm(y, y, 1.0),
             _BilinearTerm(z, z, -1.0)]
        )
        linear = _LinearExpression(
            [-5.09144, -19.96611],
            [x, y],
        )
        expr = _SumExpression([linear, quadratic])
        rule = UnivariateQuadraticRule()
        bounds = ComponentMap()
        bounds[x] = I(0, None)
        bounds[y] = I(0, None)
        bounds[z] = I(0, 4.57424778)
        bounds[expr] = I(None, -106.142171)
        result = rule.apply(expr, bounds)
        assert 5.388076 in result[x]
        assert 6.399097 in result[y]


# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import pyomo.environ as aml
from suspect.util import *
from tests.fixtures import model


def test_model_variables(model):
    assert 10 + 50 + 10 == len([_ for _ in model_variables(model)])


def test_model_constraints(model):
    model.cons1 = aml.Constraint(
        rule=lambda m: sum(m.x[i] for i in m.I) == 0
    )
    model.cons2 = aml.Constraint(
        rule=lambda m: sum(m.z[i] for i in m.I) <= 5
    )
    model.cons3 = aml.Constraint(
        model.I,
        rule=lambda m, i: -1 <= m.x[i] <= 1
    )
    model.cons4 = aml.Constraint(
        model.I, model.J,
        rule=lambda m, i, j: -20 <= m.x[i] * m.y[i, j] <= 20,
    )
    assert 1 + 1 + 10 + 50 == len([_ for _ in model_constraints(model)])


def test_model_objectives(model):
    model.obj1 = aml.Objective(expr=sum(model.x[i] for i in model.I))
    model.obj2 = aml.Objective(expr=model.x[0] - model.x[1])
    assert 2 == len([_ for _ in model_objectives(model)])



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic expression support."""
from enum import IntEnum


class ExpressionType(IntEnum):
    """Expression types supported by SUSPECT."""
    Variable = 1
    Constant = 2
    Constraint = 3
    Objective = 4
    Division = 5
    Product = 6
    Linear = 7
    Sum = 8
    Power = 9
    # While negation is technically an unary function, it's equivalent to a linear
    # expression with coefficient -1
    Negation = 10
    UnaryFunction = 11
    Quadratic = 12
    Reciprocal = 13


class UnaryFunctionType(IntEnum):
    """Unary function type supported by SUSPECT."""
    Abs = 1
    Sqrt = 2
    Exp = 3
    Log = 4
    Sin = 5
    Cos = 6
    Tan = 7
    Asin = 8
    Acos = 9
    Atan = 10



# pylint: skip-file
import pytest
from hypothesis import given, assume
import hypothesis.strategies as st
from tests.conftest import PlaceholderExpression as PE
from tests.strategies import reals
from suspect.expression import ExpressionType as ET, UnaryFunctionType as UFT
from suspect.interval import Interval as I
from suspect.convexity import Convexity
from suspect.extras.convexity.norm import L2NormRule
from suspect.math import almosteq
from mpmath import mp


@st.composite
def constants(draw):
    return PE(ET.Constant)


@st.composite
def powers(draw):
    base = PE(ET.Variable)
    expo = PE(ET.Constant, value=2.0)
    return PE(ET.Power, [base, expo])


@st.composite
def products(draw):
    x = PE(ET.Variable)
    return PE(ET.Product, [x, x])


@pytest.mark.skip('Not updated')
@given(st.lists(st.one_of(constants(), powers(), products()), min_size=1))
def test_l2_norm(children):
    rule = L2NormRule()
    sum_ = PE(ET.Sum, children)
    result = rule.checked_apply(PE(ET.UnaryFunction, [sum_], func_type=UFT.Sqrt), None)
    assert result.is_convex()



import gc
import operator as op
import xml.etree.ElementTree as ET
from contextlib import contextmanager
from functools import reduce

import pyomo.environ as aml
from pyomo.core.expr.numeric_expr import NumericConstant

NS = {'osil': 'os.optimizationservices.org'}
TYPE_TO_DOMAIN = {
    'I': aml.Integers,
    'C': aml.Reals,
    'B': aml.Binary,
}
STR_TO_SENSE = {
    'min': aml.minimize,
    'max': aml.maximize,
}


@contextmanager
def gc_disabled():
    if gc.isenabled():
        gc.disable()
        yield
        gc.enable()
    else:
        yield


def _tag_name(node):
    s = len('{os.optimizationservices.org}')
    return node.tag[s:]


def _instance_name(root):
    name = root.find('osil:instanceHeader/osil:name', NS)
    if name is not None:
        return name.text
    else:
        return None


def _instance_variables(root):
    variables = root.find('osil:instanceData/osil:variables', NS)
    for v in variables:
        attr = v.attrib
        name = attr['name']
        lb = float(attr['lb']) if 'lb' in attr else 0.0
        ub = float(attr['ub']) if 'ub' in attr else None
        type_ = attr.get('type', 'C')
        if type_ == 'D':
            raise ValueError('Unsupported semicontinuous variable')
        if type_ == 'S':
            raise ValueError('Unsupported var type S')
        bounds = (lb, ub)
        domain = TYPE_TO_DOMAIN[type_]
        value = attr.get('value', None)
        if value is not None:
            if domain == aml.Reals:
                value = float(value)
            else:
                value = int(float(value))
        yield {
            'name': name,
            'bounds': bounds,
            'domain': domain,
            'value': value,
        }


def _instance_objectives(root):
    return root.findall('osil:instanceData/osil:objectives/osil:obj', NS)


def _instance_constraints(root):
    return root.findall('osil:instanceData/osil:constraints/osil:con', NS)


def _instance_sos(root):
    return root.findall('osil:instanceData/osil:specialOrderedSets/osil:sos', NS)


def _quadratic_coefficients(root):
    qterms = root.findall('osil:instanceData/osil:quadraticCoefficients/osil:qTerm', NS)
    qterms_dict = dict()
    for qterm in qterms:
        attr = qterm.attrib
        idx = int(attr['idx'])
        idx_one = int(attr['idxOne'])
        idx_two = int(attr['idxTwo'])
        coef = float(attr['coef'])
        if idx not in qterms_dict:
            qterms_dict[idx] = list()
        qterms_dict[idx].append({
            'idx_one': idx_one,
            'idx_two': idx_two,
            'coef': coef
        })
    return qterms_dict


def _nonlinear_terms(root):
    nl_terms = root.findall('osil:instanceData/osil:nonlinearExpressions/osil:nl', NS)
    nl_terms_dict = dict()
    for nl_term in nl_terms:
        attr = nl_term.attrib
        idx = int(attr['idx'])
        assert idx not in nl_terms_dict
        nl_terms_dict[idx] = nl_term
    return nl_terms_dict


def _flatten(xs):
    return [y for x in xs for y in x]


class SparseLinearCoefficientsStorage(object):
    def __init__(self, root):
        self.root = root
        if self.root is None:
            return
        self.has_col_idx = self.root.find('osil:colIdx', NS) is not None
        self.has_row_idx = self.root.find('osil:rowIdx', NS) is not None
        if self.has_col_idx == self.has_row_idx:
            raise RuntimeError('Invalid sparse linear storage')

        if self.has_row_idx:
            raise RuntimeError('rowIdx not implemented')

        start = [
            self._expand(el)
            for el in self.root.findall('osil:start/osil:el', NS)
        ]
        self.start = _flatten(start)

        col_idx = [
            self._expand(el)
            for el in self.root.findall('osil:colIdx/osil:el', NS)
        ]
        self.col_idx = _flatten(col_idx)

        value = [
            self._expand(el, float)
            for el in self.root.findall('osil:value/osil:el', NS)
        ]
        self.value = _flatten(value)

    def row(self, i):
        if self.root is None:
            return []
        start_idx = self.start[i]
        end_idx = self.start[i+1]
        vars_with_coef = zip(
            self.value[start_idx:end_idx], self.col_idx[start_idx:end_idx]
        )
        return vars_with_coef

    def _expand(self, el, text_fun=int):
        mult = int(el.attrib.get('mult', 1))
        incr = int(el.attrib.get('incr', 0))
        start = text_fun(el.text)
        res = [start] * mult
        for i in range(1, mult):
            res[i] = res[i-1] + incr
        return res


class OsilParser(object):
    def __init__(self, filename, objective_prefix=None, constraint_prefix=None):
        self.tree = ET.parse(filename)
        self.root = self.tree.getroot()
        self.model = aml.ConcreteModel()
        self.objective_prefix = objective_prefix
        self.constraint_prefix = constraint_prefix
        self._qterms_dict = None
        self._nl_terms_dict = None

        self.indexed_vars = []

    def _v(self, i):
        vname = self.indexed_vars[i]
        return getattr(self.model, vname)

    def _quadratic_terms(self, i):
        if i not in self._qterms_dict:
            return 0
        return sum(
            c['coef'] * self._v(c['idx_one']) * self._v(c['idx_two'])
            for c in self._qterms_dict[i]
        )

    def _nonlinear_terms(self, i):
        def _eval(node):
            name = _tag_name(node)
            cs = list(node)

            if name == 'negate':
                assert len(cs) == 1
                return - _eval(cs[0])
            elif name == 'number':
                return float(node.attrib['value'])
            elif name == 'variable':
                c = float(node.attrib.get('coef', 1))
                return c * self._v(int(node.attrib['idx']))
            elif name == 'power':
                assert len(cs) == 2
                b = _eval(cs[0])
                e = _eval(cs[1])
                return b ** e
            elif name == 'square':
                assert len(cs) == 1
                return _eval(cs[0])**2
            elif name == 'sqrt':
                assert len(cs) == 1
                return aml.sqrt(_eval(cs[0]))
            elif name == 'product':
                return reduce(op.mul, [_eval(c) for c in cs])
            elif name == 'divide':
                assert len(cs) == 2
                return _eval(cs[0]) / _eval(cs[1])
            elif name == 'times':
                assert len(cs) == 2
                return _eval(cs[0]) * _eval(cs[1])
            elif name == 'plus':
                assert len(cs) == 2
                return _eval(cs[0]) + _eval(cs[1])
            elif name == 'sum':
                return sum([_eval(c) for c in cs])
            elif name == 'minus':
                assert len(cs) == 2
                return _eval(cs[0]) - _eval(cs[1])
            elif name == 'abs':
                assert len(cs) == 1
                return abs(_eval(cs[0]))
            elif name == 'exp':
                assert len(cs) == 1
                return aml.exp(_eval(cs[0]))
            elif name == 'ln':
                assert len(cs) == 1
                return aml.log(_eval(cs[0]))
            elif name == 'sin':
                assert len(cs) == 1
                return aml.sin(_eval(cs[0]))
            elif name == 'cos':
                assert len(cs) == 1
                return aml.cos(_eval(cs[0]))
            elif name == 'log10':
                assert len(cs) == 1
                return aml.log10(_eval(cs[0]))
            raise RuntimeError('unhandled tag {}'.format(name))
        if i not in self._nl_terms_dict:
            return 0.0
        nl = self._nl_terms_dict[i]
        with gc_disabled():
            children = list(nl)
            expr = _eval(children[0])
        return expr

    def _objective_linear(self, objective):
        return sum(
            float(c.text) * self._v(int(c.attrib['idx']))
            for c in objective
        ) + float(objective.attrib.get('constant', 0.0))

    def _objective_name(self, objective):
        if self.objective_prefix is None:
            return objective.attrib['name']
        else:
            return self.objective_prefix + objective.attrib['name']

    def _constraint_name(self, constraint):
        if self.constraint_prefix is None:
            return constraint.attrib['name']
        else:
            return self.constraint_prefix + constraint.attrib['name']

    def parse(self):
        sos = list(_instance_sos(self.root))
        if len(sos) != 0:
            raise NotImplementedError('The OsilParser does not support SOS constraints.')
        
        self.model.name = _instance_name(self.root)

        for var_def in _instance_variables(self.root):
            new_var = aml.Var(
                bounds=var_def['bounds'],
                domain=var_def['domain'],
                initialize=var_def['value'],
            )
            setattr(self.model, var_def['name'], new_var)
            self.indexed_vars.append(var_def['name'])

        self._qterms_dict = _quadratic_coefficients(self.root)
        self._nl_terms_dict = _nonlinear_terms(self.root)

        for i, objective in enumerate(_instance_objectives(self.root)):
            linear = self._objective_linear(objective)
            quad = self._quadratic_terms(-i-1)
            nl = self._nonlinear_terms(-i-1)
            expr = linear + quad + nl
            obj_name = self._objective_name(objective)
            sense = STR_TO_SENSE[objective.attrib['maxOrMin']]
            obj = aml.Objective(sense=sense, expr=expr)
            setattr(self.model, obj_name, obj)

        lcc = self.root.find('osil:instanceData/osil:linearConstraintCoefficients', NS)
        linear_coefficients = SparseLinearCoefficientsStorage(lcc)
        for i, constraint in enumerate(_instance_constraints(self.root)):
            linear = sum(c * self._v(v) for c, v in linear_coefficients.row(i))
            quad = self._quadratic_terms(i)
            nl = self._nonlinear_terms(i)
            expr = linear + quad + nl
            cons_name = self._constraint_name(constraint)
            lb = constraint.attrib.get('lb', None)
            ub = constraint.attrib.get('ub', None)
            if lb is not None:
                lb = NumericConstant(float(lb))
            if ub is not None:
                ub = NumericConstant(float(ub))
            # some osil files from MINLPLib have constraints without upper or lower
            # bounds (e.g., arki0011, constraint e1766)
            if lb is not None or ub is not None:
                cons = aml.Constraint(expr=aml.inequality(lb, expr, ub))
                setattr(self.model, cons_name, cons)

        return self.model


def read_osil(filename, objective_prefix=None, constraint_prefix=None):
    """Read OSiL-formatted problem into a Pyomo model.

    It's possible to optionally specify a prefix for objectives/constraints
    to avoid name clashes.

    Parameters
    ----------
    filename : str
        the OSiL file path.
    objective_prefix : str
        a prefix to give all objectives.
    constraint_prefix : str
        a prefix to give all constraints.

    Returns
    -------
    ConcreteModel
        a Pyomo concrete model.
    """
    p = OsilParser(filename, objective_prefix, constraint_prefix)
    return p.parse()



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Directed Acyclic Graph representation of an optimization problem."""

from pyomo.core.expr.numeric_expr import nonpyomo_leaf_types
from suspect.interfaces import Problem
from suspect.dag.expressions import Constraint, Objective
from suspect.dag.vertices_list import VerticesList


class ProblemDag(Problem):
    r"""The optimization problem represented as Directed Acyclic Graph (DAG).

    The vertices in the DAG are sorted by depth, defined as

    .. math::

        d(v) = \begin{cases}
          0 & \text{if } v \text{ is a variable or constant}\\
          \max\{d(u) | u \in c(v)\} & \text{otherwise}
        \end{cases}

    where :math:`c(v)` are the children of vertex :math:`v`.


    Attributes
    ----------
    name : str
        the problem name.
    variables : dict
        the problem variables.
    constraints : dict
        the problem constraints.
    objectives : dict
        the problem objectives.
    vertices : iter
        an iterator over the vertices sorted by :math:`d(v)`.
    """
    def __init__(self, name=None):
        self.name = name
        self._depths = {}
        # Keep a list of vertices that depend on a vertex
        self._parents = {}
        # The DAG vertices sorted by depth
        self._vertices = VerticesList()
        # Vertices that have no children
        self._sources = []
        # Vertices that have no parent
        self._sinks = []
        # A pointer to vertices that are variables
        self.variables = {}
        # A pointer to vertices that are constraints
        self.constraints = {}
        # A pointer to vertices that are objectives
        self.objectives = {}

    @property
    def vertices(self):
        """Return an iterator over the problem vertices."""
        return iter(self._vertices)

    def add_vertex(self, vertex):
        """Add a vertex to the DAG.

        Parameters
        ----------
        vertex : Expression
           the vertex to add.
        """
        depth = self._vertex_depth(vertex)
        self._depths[id(vertex)] = depth
        self._vertices.append(vertex, depth)
        self._add_parents(vertex)
        if _is_source_type(vertex):
            self._sources.append(vertex)

        if _is_sink_type(vertex):
            self._sinks.append(vertex)

    def _add_named(self, expr, collection):
        self.add_vertex(expr)
        collection[expr.name] = expr

    def add_variable(self, var):
        """Add a variable to the DAG.

        Parameters
        ----------
        var : Variable
            the variable.
        """
        self._add_named(var, self.variables)

    def add_constraint(self, cons):
        """Add a constraint to the DAG.

        Parameters
        ----------
        cons : Constraint
           the constraint.
        """
        self._add_named(cons, self.constraints)

    def add_objective(self, obj):
        """Add an objective to the DAG.

        Parameters
        ----------
        obj : Objective
           the objective.
        """
        self._add_named(obj, self.objectives)

    def stats(self):
        """Return statistics about the DAG.

        Returns
        -------
        dict
            A dictionary containing information about the DAG:

             * Number of vertices
             * Maximum depth
             * Number of variables
             * Number of constraints
             * Number of objectives
        """
        return {
            'num_vertices': len(self.vertices),
            'num_variables': len(self.variables),
            'num_constraints': len(self.constraints),
            'num_objectives': len(self.objectives),
        }

    def depth(self, vertex):
        if type(vertex) in nonpyomo_leaf_types:
            return 2
        if vertex.is_variable_type():
            return 1
        if vertex.is_constant():
            return 2
        return self._depths[id(vertex)]

    def _vertex_depth(self, vertex):
        assert id(vertex) not in self._depths
        if vertex.is_variable_type():
            return 1
        if vertex.is_constant():
            return 2
        depth = 0
        for arg in vertex.args:
            arg_depth = self.depth(arg)
            if arg_depth >= depth:
                depth = arg_depth + 1
        return depth

    def parents(self, vertex):
        if isinstance(vertex, (Constraint, Objective)):
            return []
        return self._parents[id(vertex)].values()

    def _add_parents(self, vertex):
        # For each of the args of vertex, add vertex as parent
        # Parents are stored in a dict so that we have both id and node
        if type(vertex) in nonpyomo_leaf_types:
            return

        if not vertex.is_expression_type():
            return

        for arg in vertex.args:
            if id(arg) not in self._parents:
                self._parents[id(arg)] = {}
            parents = self._parents[id(arg)]
            parents[id(vertex)] = vertex


def _is_source_type(node):
    is_nonpyomo_leaf = node.__class__ in nonpyomo_leaf_types
    is_variable = node.is_variable_type()
    is_constant = node.is_constant()
    return is_nonpyomo_leaf or is_variable or is_constant


def _is_sink_type(node):
    return isinstance(node, (Constraint, Objective))



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Monotonicity detection rules for division expressions."""
from suspect.monotonicity.monotonicity import Monotonicity
from suspect.monotonicity.rules.rule import MonotonicityRule
from suspect.interval import Interval


class DivisionRule(MonotonicityRule):
    """Return monotonicity of division."""
    def apply(self, expr, monotonicity, bounds):
        f, g = expr.args
        mono_f = monotonicity[f]
        mono_g = monotonicity[g]
        bound_f = bounds[f]
        bound_g = bounds[g]
        return _division_monotonicity(mono_f, bound_f, mono_g, bound_g)


class ReciprocalRule(MonotonicityRule):
    """Return monotonicity of reciprocal."""
    def apply(self, expr, monotonicity, bounds):
        g = expr.args[0]
        mono_g = monotonicity[g]
        bound_g = bounds[g]
        return _division_monotonicity(
            Monotonicity.Constant, Interval(1.0, 1.0), mono_g, bound_g
        )


def _division_monotonicity(mono_f, bound_f, mono_g, bound_g):
    # check for division by 0
    if mono_g.is_constant() and bound_g.is_zero():
        return Monotonicity.Unknown

    if mono_f.is_constant() and mono_g.is_constant():
        return Monotonicity.Constant
    elif mono_g.is_constant():
        if mono_f.is_nondecreasing() and bound_g.is_nonnegative():
            return Monotonicity.Nondecreasing
        elif mono_f.is_nonincreasing() and bound_g.is_nonpositive():
            return Monotonicity.Nondecreasing
        elif mono_f.is_nondecreasing() and bound_g.is_nonpositive():
            return Monotonicity.Nonincreasing
        elif mono_f.is_nonincreasing() and bound_g.is_nonnegative():
            return Monotonicity.Nonincreasing
        return Monotonicity.Unknown # pragma: no cover

    nondec_cond1 = (
        mono_f.is_nondecreasing() and bound_g.is_nonnegative()
    ) or (
        mono_f.is_nonincreasing() and bound_g.is_nonpositive()
    )
    nondec_cond2 = (
        bound_f.is_nonnegative() and mono_g.is_nonincreasing()
    ) or (
        bound_f.is_nonpositive() and mono_g.is_nondecreasing()
    )
    noninc_cond1 = (
        mono_f.is_nonincreasing() and bound_g.is_nonnegative()
    ) or (
        mono_f.is_nondecreasing() and bound_g.is_nonpositive()
    )
    noninc_cond2 = (
        bound_f.is_nonnegative() and mono_g.is_nondecreasing()
    ) or (
        bound_f.is_nonpositive() and mono_g.is_nonincreasing()
    )

    if nondec_cond1 and nondec_cond2:
        return Monotonicity.Nondecreasing
    elif noninc_cond1 and noninc_cond2:
        return Monotonicity.Nonincreasing
    return Monotonicity.Unknown # pragma: no cover



# pylint: skip-file
import pytest
from hypothesis import given, assume
from tests.conftest import PlaceholderExpression as PE
from tests.strategies import reals
from suspect.expression import ExpressionType as ET
from suspect.interval import Interval as I
from suspect.convexity import Convexity
from suspect.extras.convexity.fractional import FractionalRule
from suspect.math import almosteq
from mpmath import mp

MATH_PRECISION = 50
# mp.dps = MATH_PRECISION
MAX_NUM = 10**(MATH_PRECISION//2)

# TEST SETUP
#
# We want to test convexity of
#
# f(x) = (a_1 x + b_1) / (a_2 x + b_2)
#
# its second derivative is
#
# f''(x) = -2 a_2 ( (a_1 b_2 - a_2 b_1) / (a_2 x + b_2)^3 )
#
# So the sign of f''(x) depends on a_2, (a_1 b_2 - a_2 b_1) and (a_2 x + b_2)
#
# We test with denominator/denominator linear, constant and variable.

class FractionalContext:
    def __init__(self, bounds=None):
        if bounds is None:
            bounds = {}
        self._b = bounds

    def bounds(self, expr):
        return self._b[expr]


@pytest.mark.skip('Not update')
# (a_1 x) / x is linear (constant)
@given(coef=reals(allow_infinity=False))
def test_linear_over_variable(coef):
    rule = FractionalRule()
    x = PE(ET.Variable)
    num = PE(ET.Linear, [x], coefficients=[coef], constant_term=0.0)
    ctx = FractionalContext({
        x: I(0, None),
    })
    result = rule.apply(PE(ET.Division, [num, x]), ctx)
    assert result == Convexity.Linear


@pytest.mark.skip('Not update')
# (a_1 x + b_1) / x
@given(coef=reals(allow_infinity=False), const=reals(allow_infinity=False))
def test_linear_with_constant_over_variable(coef, const):
    assume(coef != 0.0)
    assume(coef < MAX_NUM and const < MAX_NUM)

    rule = FractionalRule()
    x = PE(ET.Variable)
    num = PE(ET.Linear, [x], coefficients=[coef], constant_term=const)
    ctx = FractionalContext({
        x: I(0, None),
    })
    result = rule.apply(PE(ET.Division, [num, x]), ctx)
    if almosteq(const, 0):
        expected = Convexity.Linear
    elif const > 0:
        expected = Convexity.Convex
    elif const < 0:
        expected = Convexity.Concave
    # assert result == expected



import hypothesis.strategies as st
import numpy as np
import pyomo.environ as pe
import suspect.pyomo.expressions as py5
from suspect.interval import Interval


@st.composite
def coefficients(draw, min_value=None, max_value=None):
    return draw(st.floats(
        min_value=min_value, max_value=max_value,
        allow_nan=False, allow_infinity=False,
    ))


@st.composite
def intervals(draw, allow_infinity=True):
    a = draw(reals(allow_infinity=allow_infinity))
    if np.isinf(a):
        allow_infinity=False
    b = draw(reals(allow_infinity=allow_infinity))
    lb, ub = min(a, b), max(a, b)
    return Interval(lb, ub)


@st.composite
def reals(draw, min_value=None, max_value=None, allow_infinity=True):
    if min_value is not None and max_value is not None:
        allow_infinity = False
    return draw(st.floats(
        min_value=min_value, max_value=max_value,
        allow_nan=False, allow_infinity=allow_infinity
    ))


domains = lambda: st.one_of(
    st.just(pe.Reals),
    st.just(pe.Integers),
    st.just(pe.PositiveIntegers),
    st.just(pe.Binary))


@st.composite
def variables(draw, domains=domains()):
    domain = draw(domains)
    return pe.Var(domain=domain)


constants = lambda: st.floats(min_value=-1e19, max_value=1e19, allow_nan=False, allow_infinity=False)


leaves = lambda: st.one_of(variables(), constants())


unary_function_types = st.one_of(
    st.just('cos'),
    st.just('sin'),
    st.just('tan'),
    st.just('acos'),
    st.just('asin'),
    st.just('atan'),
    st.just('exp'),
    st.just('log'),
    st.just('abs'))


@st.composite
def unary_functions(draw, child):
    ch = draw(child)
    name = draw(unary_function_types)
    if name == 'abs':
        return abs(ch)
    return getattr(pe, name)(ch)


@st.composite
def monomials(draw, vars=variables()):
    var = draw(vars)
    const = draw(constants())
    return py5.MonomialTermExpression([const, var])


@st.composite
def products(draw, arg1, arg2):
    return py5.ProductExpression([draw(arg1), draw(arg2)])


@st.composite
def sums(draw, args):
    children = draw(args)
    return py5.SumExpression(children)


def _combine_nodes(nodes):
    return st.one_of(
        unary_functions(nodes),
        products(nodes, nodes),
        sums(st.lists(nodes, min_size=2)))


def expressions(variables=variables(), max_leaves=100):
    return st.recursive(variables | monomials(variables), _combine_nodes)


@st.composite
def constraints(draw, variables=variables()):
    lb = draw(constants() | st.none())
    if lb is None:
        ub = draw(constants())
    else:
        ub = draw(constants() | st.none())

    if lb is not None and ub is not None:
        lb, ub = min(lb, ub), max(lb, ub)

    root_expr = draw(expressions(variables))

    if lb is not None:
        lb = py5.NumericConstant(lb)
    if ub is not None:
        ub = py5.NumericConstant(ub)

    expr = pe.inequality(lb, root_expr, ub)
    return pe.Constraint(expr=expr)


@st.composite
def models(draw, max_constraints=None):
    m = pe.ConcreteModel()

    vars = draw(st.lists(variables(), min_size=1))
    for i, var in enumerate(vars):
        setattr(m, 'x_{}'.format(i), var)

    one_of_vars = st.one_of([st.just(v) for v in vars])

    m.obj = pe.Objective(expr=draw(expressions(one_of_vars)))

    cons = draw(st.lists(constraints(one_of_vars), max_size=max_constraints))
    for i, con in enumerate(cons):
        setattr(m, 'c_{}'.format(i), con)

    return m



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from suspect.math import make_number
from suspect.float_hash import BTreeFloatHasher, RoundFloatHasher


@pytest.mark.skip('Only works with non-default Arb precision math')
def test_btree_float_hasher():
    hasher = BTreeFloatHasher()

    # fill hasher with some numbers
    for i in range(100):
        hasher.hash(make_number(i) / 10.0)
        hasher.hash(make_number(-i) / 5.0)

    h1 = hasher.hash(10.123)
    h2 = hasher.hash(10.123)
    h3 = hasher.hash(10.1234)
    h4 = hasher.hash(10.123000000001)

    assert h1 == h2
    assert h2 != h3
    assert h2 != h4


def test_round_float_hasher():
    hasher = RoundFloatHasher(3)
    assert hasher.hash(100.123) == hasher.hash(100.123)
    assert hasher.hash(100.123) == hasher.hash(100.123456)
    assert hasher.hash(10) == hasher.hash(10.0001)



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FBBT bounds propagation rules."""
import operator
from functools import reduce
from suspect.pyomo.expressions import nonpyomo_leaf_types
from suspect.interval import Interval
from suspect.interfaces import Rule
from suspect.expression import ExpressionType, UnaryFunctionType
from suspect.math import almosteq # pylint: disable=no-name-in-module


class VariableRule(Rule):
    """Bound propagation rule for variables."""
    def apply(self, expr, _bounds):
        return Interval(expr.lb, expr.ub)


class ConstantRule(Rule):
    """Bound propagation rule for constants."""
    def apply(self, expr, _bounds):
        if type(expr) in nonpyomo_leaf_types:
            return Interval(expr, expr)
        return Interval(expr.value, expr.value)


class ConstraintRule(Rule):
    """Bound propagation rule for constraints."""
    def apply(self, expr, bounds):
        child = expr.args[0]
        child_bounds = bounds[child]
        constraint_bounds = Interval(expr.lower_bound, expr.upper_bound)
        return constraint_bounds.intersect(child_bounds)


class ObjectiveRule(Rule):
    """Bound propagation rule for objectives."""
    def apply(self, expr, bounds):
        child = expr.args[0]
        return bounds[child]


class ProductRule(Rule):
    """Bound propagation rule for products."""
    def apply(self, expr, bounds):
        children_bounds = [bounds[child] for child in expr.args]
        return reduce(operator.mul, children_bounds, 1.0)


class QuadraticRule(Rule):
    """Bound propagation rule for quadratic."""
    def apply(self, expr, bounds):
        return sum([self._term_bounds(term, bounds) for term in expr.terms])

    def _term_bounds(self, term, bounds):
        if id(term.var1) != id(term.var2):
            return bounds[term.var1] * bounds[term.var2] * term.coefficient
        return term.coefficient * (bounds[term.var1] ** 2)


class DivisionRule(Rule):
    """Bound propagation rule for divisions."""
    def apply(self, expr, bounds):
        assert len(expr.args) == 2
        num, den = expr.args
        return bounds[num] / bounds[den]


class ReciprocalRule(Rule):
    """Bound propagation rule for divisions."""
    def apply(self, expr, bounds):
        assert len(expr.args) == 1
        child = expr.args[0]
        return 1.0 / bounds[child]


class LinearRule(Rule):
    """Bound propagation rule for linear expressions."""
    def apply(self, expr, bounds):
        children_contribution = sum(
            expr.coefficient(child) * bounds[child]
            for child in expr.args
        )
        constant_contribution = Interval(expr.constant_term, expr.constant_term)
        return children_contribution + constant_contribution


class SumRule(Rule):
    """Bound propagation rule for sum."""
    def apply(self, expr, bounds):
        return sum(bounds[child] for child in expr.args)


class PowerRule(Rule):
    """Bound propagation rule for power."""
    def apply(self, expr, bounds):
        assert len(expr.args) == 2
        _, expo = expr.args
        if type(expo) not in nonpyomo_leaf_types:
            if not expo.is_constant():
                return Interval(None, None)
            expo = expo.value

        is_even = almosteq(expo % 2, 0)
        is_positive = expo > 0
        if is_even and is_positive:
            return Interval(0, None)
        return Interval(None, None)


class NegationRule(Rule):
    """Bound propagation rule for negation."""
    def apply(self, expr, bounds):
        return -bounds[expr.args[0]]


class _UnaryFunctionRule(Rule):
    """Bound propagation rule for unary functions."""
    def apply(self, expr, bounds):
        assert len(expr.args) == 1
        child = expr.args[0]
        child_bounds = bounds[child]
        func = getattr(child_bounds, self.func_name)
        return func()


class AbsRule(_UnaryFunctionRule):
    """Return new bounds for abs."""
    func_name = 'abs'


class SqrtRule(_UnaryFunctionRule):
    """Return new bounds for sqrt."""
    func_name = 'sqrt'


class ExpRule(_UnaryFunctionRule):
    """Return new bounds for exp."""
    func_name = 'exp'


class LogRule(_UnaryFunctionRule):
    """Return new bounds for log."""
    func_name = 'log'



class SinRule(_UnaryFunctionRule):
    """Return new bounds for sin."""
    func_name = 'sin'


class CosRule(_UnaryFunctionRule):
    """Return new bounds for cos."""
    func_name = 'cos'


class TanRule(_UnaryFunctionRule):
    """Return new bounds for tan."""
    func_name = 'tan'


class AsinRule(_UnaryFunctionRule):
    """Return new bounds for asin."""
    func_name = 'asin'


class AcosRule(_UnaryFunctionRule):
    """Return new bounds for acos."""
    func_name = 'acos'


class AtanRule(_UnaryFunctionRule):
    """Return new bounds for atan."""
    func_name = 'atan'


_func_name_to_rule_map = dict()
_func_name_to_rule_map['abs'] = AbsRule()
_func_name_to_rule_map['sqrt'] = SqrtRule()
_func_name_to_rule_map['exp'] = ExpRule()
_func_name_to_rule_map['log'] = LogRule()
_func_name_to_rule_map['sin'] = SinRule()
_func_name_to_rule_map['cos'] = CosRule()
_func_name_to_rule_map['tan'] = TanRule()
_func_name_to_rule_map['asin'] = AsinRule()
_func_name_to_rule_map['acos'] = AcosRule()
_func_name_to_rule_map['atan'] = AtanRule()


class UnaryFunctionRule(Rule):
    """Bound propagation rule for unary functions."""
    def apply(self, expr, bounds):
        assert len(expr.args) == 1
        func_name = expr.getname()
        if func_name not in _func_name_to_rule_map:
            raise ValueError('Unknown function type {}'.format(func_name))
        return _func_name_to_rule_map[func_name].apply(expr, bounds)



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SUSPECT utility functions."""

from functools import wraps
import warnings
from suspect.pyomo.util import (
    model_variables,
    model_constraints,
    model_objectives,
    bounds_and_expr
)


def deprecated(new_func):
    """Mark a function as deprecated."""
    def _deprecated_decorator(func):
        @wraps(func)
        def _wrapper(*args, **kwargs):
            warnings.warn(
                '{} has been renamed to {}.'.format(
                    func.__name__,
                    new_func,
                ),
                DeprecationWarning,
            )
            return func(*args, **kwargs)
        return _wrapper
    return _deprecated_decorator



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FBBT bounds tightening rules."""
from pyomo.environ import ComponentMap
import pyomo.environ as pe
from suspect.expression import UnaryFunctionType
from suspect.interfaces import Rule, UnaryFunctionRule
from suspect.interval import Interval
from suspect.math import almosteq, isinf  # pylint: disable=no-name-in-module
from suspect.pyomo.expressions import nonpyomo_leaf_types

MAX_EXPR_CHILDREN = 1000


class ConstraintRule(Rule):
    """Return new bounds for constraint."""
    def apply(self, expr, _bounds):
        bounds = Interval(expr.lower_bound, expr.upper_bound)
        return [bounds]


class SumRule(Rule):
    """Return new bounds for sum."""
    def __init__(self, max_expr_children=None):
        if max_expr_children is None:
            max_expr_children = MAX_EXPR_CHILDREN
        self.max_expr_children = max_expr_children

    def apply(self, expr, bounds):
        if len(expr.args) > self.max_expr_children:  # pragma: no cover
            return None
        expr_bound = bounds[expr]
        children_bounds = [
            self._child_bounds(child_idx, child, expr, expr_bound, bounds)
            for child_idx, child in enumerate(expr.args)
        ]
        return children_bounds

    def _child_bounds(self, child_idx, _child, expr, expr_bound, bounds):
        siblings_bound = sum(
            bounds[c]
            for i, c in enumerate(expr.args)
            if i != child_idx
        )
        return expr_bound - siblings_bound


class LinearRule(Rule):
    """Return new bounds for linear expressions."""
    def __init__(self, max_expr_children=None):
        if max_expr_children is None:
            max_expr_children = MAX_EXPR_CHILDREN
        self.max_expr_children = max_expr_children

    def apply(self, expr, bounds):
        if expr.nargs() > self.max_expr_children:  # pragma: no cover
            return None
        expr_bound = bounds[expr]
        children_bounds = [
            bounds[child] * coef
            for coef, child in zip(expr.linear_coefs, expr.linear_vars)
        ]

        const = expr.constant
        children_bounds = [
            self._child_bounds(child_idx, coef, const, expr_bound, children_bounds)
            for child_idx, (coef, child) in enumerate(zip(expr.linear_coefs, expr.linear_vars))
        ]
        return children_bounds

    def _child_bounds(self, child_idx, coef, const, expr_bound, children_bounds):
        siblings_bound = sum(
            b for i, b in enumerate(children_bounds) if i != child_idx
        ) + const
        return (expr_bound - siblings_bound) / coef


class QuadraticRule(Rule):
    """Return new bounds for quadratic expressions."""
    def __init__(self, max_expr_children=None):
        if max_expr_children is None:
            max_expr_children = MAX_EXPR_CHILDREN
        self.max_expr_children = max_expr_children

    def apply(self, expr, bounds):
        expr_bound = bounds[expr]
        child_bounds = ComponentMap()
        if len(expr.terms) > self.max_expr_children:
            return None

        # Build bounds for all terms
        terms_bounds = [self._term_bound(t, bounds) for t in expr.terms]

        terms = expr.terms

        for term_idx, term in enumerate(terms):
            var1 = term.var1
            var2 = term.var2
            siblings_bound = sum(
                terms_bounds[i]
                for i, t in enumerate(terms) if i != term_idx
            )
            term_bound = (expr_bound - siblings_bound) / term.coefficient
            if id(var1) == id(var2):
                term_bound = term_bound.intersect(Interval(0, None))
                upper_bound = term_bound.sqrt().upper_bound
                new_bound = Interval(-upper_bound, upper_bound)

                if var1 in child_bounds:
                    existing = child_bounds[var1]
                    child_bounds[var1] = existing.intersect(new_bound)
                else:
                    child_bounds[var1] = Interval(new_bound.lower_bound, new_bound.upper_bound)

            else:
                new_bound_var1 = term_bound / bounds[var2]
                new_bound_var2 = term_bound / bounds[var1]

                if var1 in child_bounds:
                    existing = child_bounds[var1]
                    child_bounds[var1] = existing.intersect(new_bound_var1)
                else:
                    child_bounds[var1] = new_bound_var1

                if var2 in child_bounds:
                    existing = child_bounds[var2]
                    child_bounds[var2] = existing.intersect(new_bound_var2)
                else:
                    child_bounds[var2] = new_bound_var2

        return child_bounds

    def _term_bound(self, term, bounds):
        if term.var1 is term.var2:
            return term.coefficient * (bounds[term.var1] ** 2)
        return term.coefficient * bounds[term.var1] * bounds[term.var2]


class PowerRule(Rule):
    """Return new bounds for power expressions."""
    def apply(self, expr, bounds):
        base, expo = expr.args
        if type(expo) not in nonpyomo_leaf_types:
            if not expo.is_constant():
                return None
            expo = expo.value

        if not almosteq(expo, 2):
            return None

        expr_bound = bounds[expr]
        # the bound of a square number is never negative, but check anyway to
        # avoid unexpected crashes.
        if not expr_bound.is_nonnegative():
            return None

        sqrt_bound = expr_bound.sqrt()
        return [
            Interval(-sqrt_bound.upper_bound, sqrt_bound.upper_bound),
            None,
        ]


class MonomialTermRule(Rule):
    """Return new bounds for monomial term expressions."""
    def apply(self, expr, bounds):
        expr_bound = bounds[expr]
        const, expr = expr.args
        const = pe.value(const)
        if almosteq(const, 0.0):
            return None
        child_bounds = ComponentMap()
        child_bounds[expr] = expr_bound / const
        return child_bounds


class _UnaryFunctionRule(UnaryFunctionRule):
    def apply(self, expr, bounds):
        expr_bounds = bounds[expr]
        return [self._child_bounds(expr_bounds)]

    def _child_bounds(self, bounds):
        pass


class _BoundedFunctionRule(_UnaryFunctionRule):
    func_name = None

    def _child_bounds(self, bounds):
        func = getattr(bounds, self.func_name)
        return func.inverse()


class AbsRule(_BoundedFunctionRule):
    """Return new bounds for abs."""
    func_name = 'abs'


class SqrtRule(_BoundedFunctionRule):
    """Return new bounds for sqrt."""
    func_name = 'sqrt'


class ExpRule(_BoundedFunctionRule):
    """Return new bounds for exp."""
    func_name = 'exp'


class LogRule(_BoundedFunctionRule):
    """Return new bounds for log."""
    func_name = 'log'


class _UnboundedFunctionRule(_UnaryFunctionRule):
    def _child_bounds(self, bounds):
        return Interval(None, None)


class SinRule(_UnboundedFunctionRule):
    """Return new bounds for sin."""
    func_type = UnaryFunctionType.Sin


class CosRule(_UnboundedFunctionRule):
    """Return new bounds for cos."""
    func_type = UnaryFunctionType.Cos


class TanRule(_UnboundedFunctionRule):
    """Return new bounds for tan."""
    func_type = UnaryFunctionType.Tan


class AsinRule(_UnboundedFunctionRule):
    """Return new bounds for asin."""
    func_type = UnaryFunctionType.Asin


class AcosRule(_UnboundedFunctionRule):
    """Return new bounds for acos."""
    func_type = UnaryFunctionType.Acos


class AtanRule(_UnboundedFunctionRule):
    """Return new bounds for atan."""
    func_type = UnaryFunctionType.Atan


_func_name_to_rule_map = dict()
_func_name_to_rule_map['abs'] = AbsRule()
_func_name_to_rule_map['sqrt'] = SqrtRule()
_func_name_to_rule_map['exp'] = ExpRule()
_func_name_to_rule_map['log'] = LogRule()
_func_name_to_rule_map['sin'] = SinRule()
_func_name_to_rule_map['cos'] = CosRule()
_func_name_to_rule_map['tan'] = TanRule()
_func_name_to_rule_map['asin'] = AsinRule()
_func_name_to_rule_map['acos'] = AcosRule()
_func_name_to_rule_map['atan'] = AtanRule()


class UnaryFunctionRule(Rule):
    """Bound tightening rule for unary functions."""
    def apply(self, expr, bounds):
        assert len(expr.args) == 1
        func_name = expr.getname()
        if func_name not in _func_name_to_rule_map:
            raise ValueError('Unknown function type {}'.format(func_name))
        return _func_name_to_rule_map[func_name].apply(expr, bounds)



# pylint: skip-file
import pytest
from hypothesis import given, assume
import hypothesis.strategies as st
from tests.strategies import reals
from suspect.interval import Interval, EmptyInterval
from suspect.math import *


@st.composite
def intervals(draw, allow_infinity=True, lower_bound=None, upper_bound=None):
    if lower_bound is None and allow_infinity:
        lower = draw(st.one_of(
            st.none(),
            reals(max_value=upper_bound, allow_infinity=False)
        ))
    else:
        lower = draw(reals(
            min_value=lower_bound,
            max_value=upper_bound,
            allow_infinity=False
        ))

    if upper_bound is None and allow_infinity:
        upper = draw(st.one_of(
            st.none(),
            reals(min_value=lower, allow_infinity=False)
        ))
    else:
        upper = draw(reals(
            min_value=lower,
            max_value=upper_bound,
            allow_infinity=False,
        ))
    return Interval(lower, upper)


class TestAddition(object):
    @given(intervals())
    def test_addition_with_zero(self, bound):
        zero_bound = Interval.zero()
        assert bound + zero_bound == bound

    @given(intervals(), intervals())
    def test_commutative_property(self, a, b):
        assert a + b == b + a

    @given(intervals(), intervals(lower_bound=0))
    def test_addition_with_positive_bound(self, a, b):
        c = a + b
        s = Interval(a.lower_bound, None)
        assert c in s

    @given(intervals(), intervals(upper_bound=0))
    def test_addition_with_negative_bound(self, a, b):
        c = a + b
        s = Interval(None, a.upper_bound)
        assert c in s


class TestSubtraction(object):
    @given(intervals())
    def test_negation(self, a):
        assert (-(-a)) == a

    @given(intervals())
    def test_subtraction_with_zero(self, a):
        assert a - 0 == a

    @given(
        intervals(),
        intervals(lower_bound=0),
    )
    def test_subtraction_with_positive_bound(self, a, b):
        c = a - b
        s = Interval(None, a.upper_bound)
        assert c in s

    @given(
        intervals(),
        intervals(upper_bound=0),
    )
    def test_subtraction_with_negative_bound(self, a, b):
        c = a - b
        s = Interval(a.lower_bound, None)
        assert c in s


class TestMultiplication(object):
    @given(intervals())
    def test_multiplication_with_zero(self, a):
        assert (a * 0).is_zero()

    @given(intervals())
    def test_multiplication_with_one(self, a):
        assert a * 1.0 == a

    @given(intervals(), intervals())
    def test_commutative_property(self, a, b):
        c = a * b
        d = b * a
        assert c == d

    @given(intervals(lower_bound=0), intervals(lower_bound=0))
    def test_multiplication_positive_with_positive(self, a, b):
        c = a * b
        assert c.is_nonnegative()

    @given(intervals(lower_bound=0), intervals(upper_bound=0))
    def test_multiplication_positive_with_negative(self, a, b):
        c = a * b
        assert c.is_nonpositive()


class TestDivision(object):
    @given(intervals(), intervals())
    def test_division_by_zero(self, a, b):
        assume(b.lower_bound < 0 and b.upper_bound > 0)
        c = a / b
        assert c == Interval(None, None)

    @given(intervals())
    def test_division_by_one(self, a):
        assert a / 1.0 == a

    @given(intervals(lower_bound=0), intervals(lower_bound=1))
    def test_division_positive_with_positive(self, a, b):
        c = a / b
        assert c.is_nonnegative()

    @given(intervals(lower_bound=0), intervals(upper_bound=-1))
    def test_division_positive_with_negative(self, a, b):
        c = a / b
        assert c.is_nonpositive()


class TestContains(object):
    @given(intervals())
    def test_bound_contains_itself(self, a):
        assert a in a

    @given(intervals(allow_infinity=False))
    def test_bound_contains_midpoint(self, a):
        assume((a.upper_bound - a.lower_bound) != inf)
        m = a.lower_bound + (a.upper_bound - a.lower_bound) / 2.0
        assert m in a

    @given(reals())
    def test_infinity_bound_contains_everything(self, n):
        a = Interval(None, None)
        assert n in a


class TestIntersect(object):
    @given(intervals(), intervals())
    def test_bound_is_tighter(self, a, b):
        assume(a.upper_bound > b.lower_bound)
        assume(a.lower_bound < b.upper_bound)
        c = a.intersect(b)
        assert c.lower_bound >= min(a.lower_bound, b.lower_bound)
        assert c.upper_bound <= max(a.upper_bound, b.upper_bound)


class TestSize(object):
    @given(reals())
    def test_infinite_lower_bound_greater_than_everything(self, f):
        b = Interval(None, 0)
        assert almostgte(b.size(), f)

    @given(reals())
    def test_infinite_upper_bound_greater_than_everything(self, f):
        b = Interval(0, None)
        assert almostgte(b.size(), f)

    @given(reals())
    def test_infinite_bounds_greater_than_everything(self, f):
        b = Interval(None, None)
        assert almostgte(b.size(), f)

    @given(
        intervals(allow_infinity=False, upper_bound=1e10),
        reals(allow_infinity=False, min_value=1.0, max_value=1e20)
    )
    def test_finite_bounds(self, a, f):
        b = a * f
        assert almostlte(a.size(), b.size())

    @given(
        intervals(allow_infinity=False),
        reals(allow_infinity=False, min_value=0.0, max_value=1.0)
    )
    def test_finite_bounds_1(self, a, f):
        b = a * f
        assert almostgte(a.size(), b.size())


class TestAbs(object):
    @pytest.mark.parametrize('it,expected', [
        (Interval(1, 2), Interval(1, 2)),
        (Interval(-3, -2), Interval(2, 3)),
        (Interval(-3, 2), Interval(0, 3)),
        (Interval(-2, 3), Interval(0, 3)),
    ])
    def test_abs(self, it, expected):
        assert expected == abs(it)


class TestPower(object):
    @pytest.mark.parametrize('base,p,expected', [
        (Interval(2, 3), 3, Interval(8, 27)),
        (Interval(2, 3), 4, Interval(16, 81)),
        (Interval(-3, 2), 3, Interval(-27, 8)),
        (Interval(-3, 2), 4, Interval(0, 81)),
        (Interval(-3, -2), 3, Interval(-27, -8)),
        (Interval(-3, -2), 4, Interval(16, 81)),

        (Interval(2, 4), -3, Interval(1.0/64.0, 1.0/8.0)),
        (Interval(2, 4), -4, Interval(1.0/256.0, 1.0/16.0)),
        (Interval(-4, -2), -3, Interval(-1.0/8.0, -1.0/64.0)),
        (Interval(-4, -2), -4, Interval(1.0/256.0, 1.0/16.0)),

        (Interval(2, 3), 0, Interval(1, 1)),
        (Interval(-3, 2), 0, Interval(1, 1)),
        (Interval(-3, -2), 0, Interval(1, 1)),
    ])
    def test_power(self, base, p, expected):
        result = base ** p
        assert expected == result



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convexity detection rules for linear and sum expressions."""
from suspect.convexity.convexity import Convexity
from suspect.convexity.rules.rule import ConvexityRule


class LinearRule(ConvexityRule):
    """Return convexity of linear expression."""
    def apply(self, expr, convexity, mono, bounds):
        cvxs = [
            _adjust_convexity(convexity[child], expr.coefficient(child))
            for child in expr.args
        ]
        return _combine_convexities(cvxs)


class SumRule(ConvexityRule):
    """Return convexity of sum expression."""
    def apply(self, expr, convexity, _mono, _bounds):
        cvxs = [convexity[child] for child in expr.args]
        return _combine_convexities(cvxs)


def _adjust_convexity(cvx, coef):
    if cvx.is_unknown() or cvx.is_linear():
        return cvx

    if coef > 0:
        return cvx
    elif coef < 0:
        return cvx.negate()

    # if coef == 0, it's a constant with value 0.0
    return Convexity.Linear


def _combine_convexities(cvxs):
    all_linear = all([c.is_linear() for c in cvxs])
    all_convex = all([c.is_convex() for c in cvxs])
    all_concave = all([c.is_concave() for c in cvxs])

    if all_linear:
        return Convexity.Linear
    elif all_convex:
        return Convexity.Convex
    elif all_concave:
        return Convexity.Concave

    return Convexity.Unknown



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for visitors applying the given rules."""
import abc


class Visitor(metaclass=abc.ABCMeta):
    """Base class for visitors applying rules to DAG vertices."""
    needs_matching_rules = True

    @abc.abstractmethod
    def visit_expression(self, expr, ctx, *args):
        """Apply rule to ``expr``.

        Parameters
        ----------
        expr : expression
            the expression being visited
        ctx : dict-like
            the current ctx
        *args : dict-like list
            additional ctx

        Returns
        -------
        bool
            Whether a rule was matched
        value
            The value that will be passed to ``handle_result``"""
        pass

    @abc.abstractmethod
    def handle_result(self, expr, result, ctx):
        """Handle ``result`` obtained applying rule to ``expr``."""
        pass

    def visit(self, expr, ctx, *args):
        """Apply registered rule to ``expr``."""
        return self._visit_expression(expr, ctx, *args)

    def _handle_result(self, expr, result, ctx):
        return self.handle_result(expr, result, ctx)

    def _visit_expression(self, expr, ctx, *args):
        matched, result = self.visit_expression(expr, ctx, *args)
        if not matched:
            if self.needs_matching_rules:
                raise RuntimeError('visiting expression with no rule associated.')
            return False
        return self._handle_result(expr, result, ctx)


class ForwardVisitor(Visitor): # pylint: disable=abstract-method
    """Visitor when visiting DAG forward."""
    pass


class BackwardVisitor(Visitor): # pylint: disable=abstract-method
    """Visitor when visiting DAG backward."""
    def _handle_result(self, expr, result, ctx):
        if result is None:
            # if we allow matching rules, then continue iteration
            return not self.needs_matching_rules
        any_change = False
        if isinstance(result, list):
            if len(result) != len(expr.args):
                raise ValueError(
                    'Result must be a list with same length as expr.args'
                )
            for child, value in zip(expr.args, result):
                any_change |= self.handle_result(child, value, ctx)
            return any_change
        # Visitor returned a dict
        for arg, value in result.items():
            any_change |= self.handle_result(arg, value, ctx)
        return any_change



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: skip-file
from setuptools import setup, find_packages
from pathlib import Path

project_root = Path(__file__).resolve().parent

about = {}
version_path = project_root / 'suspect' / '__version__.py'
with version_path.open() as f:
    exec(f.read(), about)

with open('README.rst') as f:
    readme = f.read()

with open('CHANGELOG.rst') as f:
    changelog = f.read()

setup(
    name='cog-suspect',
    author=about['__author__'],
    author_email=about['__author_email__'],
    license=about['__license__'],
    version=about['__version__'],
    url=about['__url__'],
    description=about['__description__'],
    long_description=readme + '\n\n' + changelog,
    packages=find_packages(exclude=['tests.*', 'tests']),
    entry_points={
        'suspect.convexity_detection': [
            'perspective=suspect.extras.convexity:PerspectiveFunctionConvexityDetector',
            'l2norm=suspect.extras.convexity:L2NormConvexityDetector',
            'quadratic=suspect.extras.convexity:QuadraticFormConvexityDetector',
            'fractional=suspect.extras.convexity:FractionalConvexityDetector',
        ]
    },
    scripts=[
        'scripts/model_summary.py',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Mathematics',
    ],
    install_requires=[
        'pyomo>=5.6',
        'numpy>=1.14',
        'mpmath>=1.0',
        'boto3>=1.7.4',
    ],
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'pytest-cov', 'hypothesis', 'pytest-rerunfailures', 'sympy'],
)



from numbers import Number

import pyomo.environ as pyo
from pyomo.repn.standard_repn import generate_standard_repn
from pyomo.core.expr.numeric_expr import (
    nonpyomo_leaf_types,
    NumericConstant,
    SumExpression,
    LinearExpression,
)
from pyomo.core.expr.visitor import ExpressionValueVisitor
from pyomo.core.kernel.component_map import ComponentMap

from suspect.float_hash import BTreeFloatHasher
from suspect.pyomo.expr_dict import ExpressionDict
from suspect.pyomo.quadratic import QuadraticExpression


def create_connected_model(model, active=True, connect_max_linear_children=50, connect_max_quadratic_children=100):
    connected = model.clone()

    model_to_connected_map = ComponentMap()
    components = ExpressionDict(float_hasher=BTreeFloatHasher())

    for var in model.component_data_objects(pyo.Var, active=active, sort=True, descend_into=True):
        connected_var = connected.find_component(var.getname(fully_qualified=True))
        model_to_connected_map[var] = connected_var
        components[connected_var] = connected_var

    for constraint in model.component_data_objects(pyo.Constraint, active=active, sort=True, descend_into=True):
        connected_constraint = connected.find_component(constraint.getname(fully_qualified=True))
        model_to_connected_map[constraint] = connected_constraint

    for objective in model.component_data_objects(pyo.Objective, active=active, sort=True, descend_into=True):
        connected_objective = connected.find_component(objective.getname(fully_qualified=True))
        model_to_connected_map[objective] = connected_objective

    convert_visitor = _ConvertExpressionVisitor(
        components, model_to_connected_map, connect_max_linear_children, connect_max_quadratic_children
    )

    for constraint in connected.component_data_objects(pyo.Constraint, active=active, sort=True, descend_into=True):
        new_body = convert_visitor.dfs_postorder_stack(constraint.body)
        constraint._body = new_body

    for objective in connected.component_data_objects(pyo.Objective, active=active, sort=True, descend_into=True):
        new_body = convert_visitor.dfs_postorder_stack(objective.expr)
        objective._expr = new_body

    return connected, model_to_connected_map


class _ConvertExpressionVisitor(ExpressionValueVisitor):
    def __init__(self, memo, component_map, connect_max_linear_children, connect_max_quadratic_children):
        self.memo = memo
        self.component_map = component_map
        self.connect_max_linear_children = connect_max_linear_children
        self.connect_max_quadratic_children = connect_max_quadratic_children

    def visiting_potential_leaf(self, node):
        if node.__class__ in nonpyomo_leaf_types:
            expr = NumericConstant(float(node))
            expr = self.set(expr, expr)
            return True, expr

        if node.is_constant():
            expr = self.set(node, node)
            return True, expr

        if node.is_variable_type():
            var = self.get(node)
            assert var is not None
            return True, var

        if _is_quadratic_expression(node) and node.nargs() > self.connect_max_quadratic_children:
            return True, node

        if _is_linear_expression(node) and node.nargs() > self.connect_max_linear_children:
            return True, node

        return False, None

    def visit(self, node, values):
        new_expr = self.get(node)
        if new_expr is not None:
            self.set(node, new_expr)
            return new_expr

        if _is_quadratic_expression(node):
            new_expr = _convert_quadratic_expression(node)
        else:
            new_expr = node.create_node_with_local_data(tuple(values))
        self.set(node, new_expr)
        return new_expr

    def get(self, expr):
        if isinstance(expr, Number):
            const = NumericConstant(expr)
            return self.get(const)
        else:
            return self.memo[expr]

    def set(self, expr, new_expr):
        self.component_map[expr] = new_expr
        if self.memo.get(expr) is not None:
            return self.memo[expr]
        self.memo[expr] = new_expr
        return new_expr


def _is_quadratic_expression(node):
    return type(node) == SumExpression and node.polynomial_degree() == 2 and node.nargs() > 1


def _is_linear_expression(node):
    return (
        type(node) in (SumExpression, LinearExpression)
        and node.polynomial_degree() == 1
    )


def _convert_quadratic_expression(expr):
    # Check if there is any non bilinear term
    repn_result = generate_standard_repn(expr, compute_values=False)
    assert repn_result.quadratic_vars
    assert not repn_result.nonlinear_expr
    quadratic_expr = 0.0

    for (v1, v2), coeff in zip(repn_result.quadratic_vars, repn_result.quadratic_coefs):
        quadratic_expr += (v1 * v2) * float(coeff)

    if not repn_result.linear_vars:
        return QuadraticExpression(quadratic_expr)

    linear_expr = repn_result.constant

    for linear_var, linear_coeff in zip(repn_result.linear_vars, repn_result.linear_coefs):
        linear_expr += linear_var * float(linear_coeff)

    quadratic = QuadraticExpression(quadratic_expr)
    return SumExpression([quadratic, linear_expr])
